from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Text, Toplevel

if typing.TYPE_CHECKING:
    from biscuit.common.textindex import TextIndex
    from src import App

    from .text import Text


class Diagnostic(Toplevel):
    """Diagnostic tooltip for the editor

    Shows the diagnostic message for the current word under the cursor."""

    def __init__(self, master: App, bd: int = 1, *args, **kw) -> None:
        super().__init__(master, *args, **kw)
        self.overrideredirect(True)
        self.config(bg=self.base.theme.border, bd=bd)

        self.label = tk.Label(
            self,
            padx=6,
            font=self.base.settings.font,
            anchor=tk.W,
            justify=tk.LEFT,
            **self.base.theme.editors.hover.text,
        )
        self.label.pack(fill=tk.BOTH, expand=True)

        self.withdraw()

    def update_position(self, pos: str, tab: Text):
        tab.update_idletasks()

        pos_x, pos_y = tab.winfo_rootx(), tab.winfo_rooty()

        pos = tab.index(str(pos) + " wordstart")
        bbox = tab.bbox(pos)
        if not bbox:
            return

        bbx_x, bbx_y, _, height = bbox
        self.geometry("+{}+{}".format(pos_x + bbx_x - 1, pos_y + bbx_y + height))

    def show(self, tab: Text, pos: str, response: str, severity: int) -> None:
        if not response:
            return

        self.label.config(text=response)
        self.label.pack(fill=tk.BOTH, pady=(1, 0) if response else 0)

        self.update()
        self.deiconify()
        self.update_position(pos, tab)

    def hide(self, *_) -> None:
        self.withdraw()



from __future__ import annotations

import os
import platform
import tkinter as tk
import tkinter.filedialog as filedialog
import typing
import webbrowser as web
from tkinter import messagebox
from tkinter.filedialog import asksaveasfilename

from biscuit.common.classdrill import *

if typing.TYPE_CHECKING:
    from .app import App


class Commands:
    """Commands that can be triggered by the user.

    This class contains all the commands that can be triggered by the user.
    All the methods that are not decorated with `@command_palette_ignore` are exported to
    the command palette and can be triggered by the user.

    The `@command_palette_ignore` decorator is used to ignore a method from being exported to the command palette.
    """

    def __init__(self, base: App) -> None:
        self.base: App = base
        self.count = 1
        self.maximized = False
        self.minimized = False
        self.previous_pos = None

    def new_file(self, *_) -> None:
        self.base.open_editor(f"Untitled-{self.count}", exists=False)
        self.count += 1

    def new_window(self, *_) -> None:
        self.base.open_new_window()

    def open_empty_editor(self, *_) -> None:
        self.new_file()

    def open_file(self, *_) -> None:
        path = filedialog.askopenfilename()
        if not path or not os.path.isfile(path):
            return
        self.base.open_editor(path)
        self.base.history.register_file_history(path)

    def open_directory(self, *_) -> None:
        path = filedialog.askdirectory()
        if not path or not os.path.isdir(path):
            return
        self.base.open_directory(path)
        self.base.history.register_folder_history(path)

    def show_recent_files(self, *_):
        self.base.palette.show("recentf:")

    def show_recent_folders(self, *_):
        self.base.palette.show("recentd:")

    def restore_recent_session(self, *_) -> None:
        self.base.sessions.restore_session()

    def restore_last_closed_editor(self, *_) -> None:
        self.base.editorsmanager.restore_last_closed_editor()

    def save_file(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content:
                if not editor.content.exists:
                    return self.save_file_as()
                if editor.content.editable:
                    editor.save()

    def save_file_as(self, *_) -> None:
        # TODO set initial filename to a range of text inside the editor
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                if path := asksaveasfilename(
                    title="Save As...",
                    initialfile=(editor.filename or "Untitled.txt"),
                ):
                    editor.save(path)

    def save_all(self, *_) -> None:
        for editor in self.base.editorsmanager.active_editors:
            if editor.content:
                if not editor.content.exists:
                    if path := asksaveasfilename(
                        title="Save As...",
                        initialfile=(editor.filename or "Untitled.txt"),
                    ):
                        return editor.save(path)
                if editor.content.editable:
                    editor.save()

    def save_workspace_as(self, *_) -> None:
        if path := asksaveasfilename(
            title="Save Workspace As...",
            initialfile="workspace.toml",
            filetypes=[(".toml")],
        ):
            self.base.workspaces.save(path)

    def add_folder_to_workspace(self, *_) -> None:
        path = filedialog.askdirectory()
        if not path or not os.path.isdir(path):
            return
        self.base.workspaces.add_dir(path)

    def open_workspace(self, *_) -> None:
        path = filedialog.askopenfilename()
        if not path or not os.path.isfile(path):
            return
        self.base.workspaces.load(path)

    def close_workspace(self, *_) -> None:
        self.base.workspaces.close()

    def open_settings(self, *_) -> None:
        self.base.open_settings()

    def close_editor(self, *_) -> None:
        self.base.close_active_editor()

    def close_all_editors(self, *_) -> None:
        self.base.editorsmanager.delete_all_editors()

    def close_folder(self, *_) -> None:
        self.base.close_active_directory()

    def quit_biscuit(self, *_) -> None:
        self.base.on_close_app()
        # self.base.destroy()

    def split_editor(self, *_) -> None:
        self.base.editorsmanager.split_editor()

    def change_tab_forward(self, *_) -> None:
        self.base.editorsbar.change_tab_forward()
        return "break"

    def change_tab_back(self, *_) -> None:
        self.base.editorsbar.change_tab_back()
        return "break"

    def maximize_biscuit(self, *_) -> None:
        match platform.system():
            case "Windows" | "Darwin":
                self.base.wm_state("normal" if self.maximized else "zoomed")
            # TODO windows specific maximizing
            # case "Windows":
            #     from ctypes import windll
            #     if not self.maximized:
            #         hwnd = windll.user32.GetParent(self.base.winfo_id())
            #         SWP_SHOWWINDOW = 0x40
            #         windll.user32.SetWindowPos(hwnd, 0, 0, 0, int(self.base.winfo_screenwidth()), int(self.base.winfo_screenheight()-48),SWP_SHOWWINDOW)
            #     else:
            #         hwnd = windll.user32.GetParent(self.base.winfo_id())
            #         SWP_SHOWWINDOW = 0x40
            #         windll.user32.SetWindowPos(hwnd, 0, self.previous_pos[0], self.previous_pos[1], int(self.base.minsize()[0]), int(self.base.minsize()[1]), SWP_SHOWWINDOW)
            case _:
                self.base.wm_attributes("-zoomed", self.maximized)

        self.maximized = not self.maximized

    def minimize_biscuit(self, *_) -> None:
        self.base.update_idletasks()

        if platform.system() == "Windows":
            from ctypes import windll

            try:
                hwnd = windll.user32.GetParent(self.base.winfo_id())
                windll.user32.ShowWindow(hwnd, 6)
            except Exception as e:
                print(e)
                self.base.withdraw()
        else:
            self.base.withdraw()
            self.base.notifications.hide()

        self.minimized = True

    @command_palette_ignore
    def window_mapped(self, *_) -> None:
        self.base.update_idletasks()
        if self.minimized:
            self.base.deiconify()
            self.minimized = False

    def toggle_wordwrap(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                self.base.wrap_words = not self.base.wrap_words
                editor.content.text.refresh_wrap()

    def toggle_block_cursor(self, *_) -> None:
        self.base.block_cursor = not self.base.block_cursor
        if e := self.base.editorsmanager.active_editor:
            if e.content and e.content.editable:
                e.content.text.set_block_cursor(self.base.block_cursor)

    def toggle_relative_line_numbering(self, *_) -> None:
        self.base.relative_line_numbers = not self.base.relative_line_numbers
        for e in self.base.editorsmanager.active_editors:
            if e.content and e.content.editable:
                e.content.text.toggle_relative_numbering()

    def undo(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.edit_undo()

    def redo(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.edit_redo()

    def cut(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.cut()

    def copy(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.copy()

    def paste(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.paste()

    def toggle_comment(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.toggle_comment()

    def find_symbol(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.open_find_replace()

    def replace_symbol(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.open_find_replace()

    def select_all(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.tag_add("sel", "1.0", "end")

    def select_line(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.select_line(editor.content.text.index(tk.INSERT))

    def delete_line(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_delete_line()

    def copy_line_up(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_copy_line_up()

    def copy_line_down(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_copy_line_down()

    def move_line_up(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_move_line_up()

    def move_line_down(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_move_line_down()

    def duplicate_selection(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.event_duplicate_selection()

    def go_to_symbol_definition(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.request_definition(from_menu=True)

    def find_symbol_references(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.request_references(from_menu=True)

    def rename_symbol(self, *_) -> None:
        if editor := self.base.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.request_rename()

    def restart_extension_server(self, *_) -> None:
        self.base.extensions_manager.restart_server()

    def debugger_toggle_pause(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.continue_pause()

    def debugger_step_over(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.step_over()

    def debugger_step_into(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.step_in()

    def debugger_step_out(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.step_out()

    def debugger_restart(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.restart()

    def debugger_stop(self, *_):
        if self.base.debugger_manager.latest:
            self.base.debugger_manager.latest.stop()

    def toggle_sidebar(self, *_) -> None:
        self.base.sidebar.toggle()

    def toggle_secondary_sidebar(self, *_) -> None:
        self.base.secondary_sidebar.toggle()

    def show_directory_tree(self, *_) -> None:
        # TODO: some renaming needed
        self.base.sidebar.show_explorer()

    def show_outline(self, *_) -> None:
        self.base.secondary_sidebar.show_outline()

    def show_search(self, *_) -> None:
        self.base.sidebar.show_search()

    def show_source_control(self, *_) -> None:
        self.base.secondary_sidebar.show_source_control()

    def show_debugger(self, *_) -> None:
        self.base.sidebar.show_debug()

    def show_extensions(self, *_) -> None:
        self.base.secondary_sidebar.show_extensions()

    def show_assistant(self, *_) -> None:
        self.base.secondary_sidebar.show_ai()

    def show_outline(self, *_) -> None:
        self.base.secondary_sidebar.show_outline()

    def show_github(self, *_) -> None:
        self.base.secondary_sidebar.show_github()

    def show_terminal(self, *_) -> None:
        self.base.panel.show_terminal()

    def show_logs(self, *_) -> None:
        self.base.panel.show_logs()

    def show_welcome(self, *_) -> None:
        self.base.editorsmanager.add_welcome()

    def show_command_palette(self, *_) -> None:
        self.base.palette.show(">")

    def show_symbol_palette(self, *_) -> None:
        self.base.palette.show("@")

    def show_active_editors(self, *_) -> None:
        self.base.palette.show("active:")

    def change_language_mode(self, *_) -> None:
        self.base.palette.show("language:")

    def change_end_of_line_character(self, *_) -> None:
        self.base.palette.show("eol:")

    def change_encoding(self, *_) -> None:
        self.base.palette.show("encoding:")

    def change_indentation_level(self, *_) -> None:
        self.base.palette.show("indent:")

    def clone_git_repository(self, *_) -> None:
        self.base.palette.show("clone:")

    def goto_line_column(self, *_) -> None:
        self.base.palette.show(":")

    def change_git_branch(self, *_) -> None:
        self.base.palette.show("branch:")

    @command_palette_ignore
    def show_run_config_palette(self, command) -> None:
        self.base.palette.show("runconf:", command)

    def search_google(self, *_) -> None:
        self.base.palette.show("google:")

    def configure_run_command(self, *_) -> None:
        self.base.palette.show("runconf:")

    def search_github_issues(self, *_) -> None:
        self.base.palette.show("issue:")

    def search_github_prs(self, *_) -> None:
        self.base.palette.show("pr:")

    def search_files(self, *_) -> None:
        self.base.palette.show()

    def open_biscuit_documentation(self, *_) -> None:
        web.open("https://tomlin7.github.io/biscuit/")

    def open_biscuit_release_notes(self, *_) -> None:
        web.open("https://github.com/tomlin7/biscuit/blob/main/CHANGELOG.md")

    def report_bug(self, *_) -> None:
        web.open(
            "https://github.com/tomlin7/biscuit/issues/new?assignees=tomlin7&labels=bug&projects=&template=bug_report.md"
        )

    def request_feature(self, *_) -> None:
        web.open(
            "https://github.com/tomlin7/biscuit/issues/new?assignees=tomlin7&labels=enhancement&projects=&template=feature_request.md"
        )

    def open_biscuit_code_of_conduct(self, *_) -> None:
        web.open("https://github.com/tomlin7/biscuit/blob/main/CODE_OF_CONDUCT.md")

    def view_biscuit_licenses(self, *_) -> None:
        web.open("https://github.com/tomlin7/biscuit/blob/main/LICENSE.md")

    def show_about(self, *_) -> None:
        messagebox.showinfo("Biscuit", str(self.base.system))
        self.base.logger.info(str(self.base.system))



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common import Frame, Menu
from biscuit.common.icons import Icons

from .actionbutton import ActionButton
from .menubutton import ActionMenuButton

if typing.TYPE_CHECKING:
    from biscuit.views import SideBarView

    from ..sidebar import SideBar
    from .statusbar import Statusbar


class ActivityBar(Frame):
    def __init__(self, master: Statusbar, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.layout.sidebar.actionbar)

        self.buttons: list[ActionButton] = []
        self.active_button: ActionButton = None

        self.menus = []
        # self.add_menus()

    def attach_sidebar(self, sidebar: SideBar) -> None:
        self.sidebar = sidebar

    def add_view(self, view: SideBarView) -> None:
        btn = ActionButton(self, view)
        btn.pack(side=tk.LEFT)
        self.buttons.append(btn)

    def toggle_first_slot(self) -> None:
        self.buttons[0].toggle()

    def set_active_slot(self, selected_button: ActionButton) -> None:
        self.active_button = selected_button
        for button in self.buttons:
            if button != selected_button:
                button.disable()

    def add_menus(self) -> None:
        self.add_settings_menu()

    def add_settings_menu(self) -> None:
        settings_menu = self.add_menu(Icons.SETTINGS_GEAR, "manage")
        settings_menu.add_command(
            "Command Palette", lambda *_: self.base.palette.show(">")
        )
        settings_menu.add_separator()
        settings_menu.add_command("Settings", self.base.commands.open_settings)

    def add_menu(self, icon: str, text: str) -> Menu:
        menu_btn = ActionMenuButton(self, icon, text)
        menu_btn.pack(side=tk.LEFT)
        self.menus.append(menu_btn.menu)

        return menu_btn.menu

    def close_all_menus(self, *_) -> None:
        for menu in self.menus:
            menu.hide()

    def switch_menu(self, menu: Menu) -> None:
        active = False
        for i in self.menus:
            if i.active:
                active = True
            if i != menu:
                i.hide()

        if active:
            menu.show()



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Menubutton

if typing.TYPE_CHECKING:
    from biscuit.views import PanelView

    from .panelbar import PanelBar


class Tab(Menubutton):
    """Tab of panel bar
    Panel views are attached to tabs. Tabs are displayed in the panel bar."""

    def __init__(self, master: PanelBar, view: PanelView, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.view = view
        self.selected = False

        self.config(
            text=view.__class__.__name__,
            padx=5,
            pady=5,
            font=self.base.settings.uifont,
            **self.base.theme.layout.content.editors.bar.tab,
        )

        self.bind("<Button-1>", self.select)

    def deselect(self, *_) -> None:
        if self.selected:
            self.view.grid_remove()
            self.config(
                fg=self.base.theme.layout.content.editors.bar.tab.foreground,
                bg=self.base.theme.layout.content.editors.bar.tab.background,
            )
            self.selected = False

    def select(self, *_) -> None:
        if not self.selected:
            self.master.set_active_tab(self)
            self.view.grid(column=0, row=1, sticky=tk.NSEW)
            self.config(
                fg=self.base.theme.layout.content.editors.bar.tab.selectedforeground,
                bg=self.base.theme.layout.content.editors.bar.tab.selectedbackground,
            )
            self.selected = True



from .actionset import ActionSet
from .classdrill import *
from .fixedstack import FixedSizeStack
from .games import *
from .helpers import *
from .icons import Icons
from .io import IO
from .menu import Checkable, Command, Dropdown, Menu
from .notifications import Notification, Notifications
from .open_editors import OpenEditors
from .palette import Palette
from .sysinfo import SysInfo
from .text_editor_menu import TextEditorContextMenu
from .textutils import *



from __future__ import annotations

import typing

import click

from biscuit import get_app_instance

if typing.TYPE_CHECKING:
    from biscuit import App


@click.group(invoke_without_command=True)
@click.help_option("-h", "--help")
def ext():
    """Commands for managing biscuit extensions

    This command group allows you to manage biscuit extensions.

    Example:
        biscuit ext list \n
        biscuit ext install extension_name \n
        biscuit ext uninstall extension_name
    """
    ...


@ext.result_callback()
@click.pass_context
def process_extension_commands(ctx, processors):
    if ctx.invoked_subcommand is None:
        # show help
        click.echo(ctx.get_help())
        exit()

    app = get_app_instance()
    app.withdraw()
    print("\nLoading extensions...")

    while not app.extensions_manager.extensions_loaded:
        ...

    if processors:
        if isinstance(processors, list):
            for processor in processors:
                processor(app)
        else:
            processors(app)

    exit()


@ext.command("list")
@click.option("-u", "--user", help="Filter by user")
@click.option("-i", "--installed", is_flag=True, help="Show installed extensions")
def list_ext(user, installed) -> typing.Callable[[App], typing.List[str]]:
    """List all extensions or installed or filter by user

    Example:
        biscuit ext list \n
        biscuit ext list -u user \n
        biscuit ext list -i

    Args:
        user (str): Filter by user
        installed (bool): Show installed extensions
    """

    if user:
        click.echo(f"Listing extensions by {user}\n")

        def f(app: App, user=user) -> None:
            for i, data in enumerate(
                app.extensions_manager.list_extensions_by_user(user)
            ):
                click.echo(f"[{i}] {data[0]}: " + data[1][-1])

        return f

    elif installed:
        click.echo("Listing installed extensions\n")

        def f(app: App) -> None:
            for i, data in enumerate(
                app.extensions_manager.list_installed_extensions()
            ):
                click.echo(
                    f"[{i}] {data[0]}: " + ", ".join(data[1]) if data[1] else " ... "
                )

        return f
    else:
        click.echo("Listing all extensions\n")

        def f(app: App) -> None:
            for i, data in enumerate(app.extensions_manager.list_all_extensions()):
                click.echo(f"[{i}] {data[0]}: " + ", ".join(data[1][1:]))

        return f


@ext.command()
@click.argument("name")
def info(name: str) -> typing.Callable[[App], None]:
    """Show information about an extension by name

    Example:
        biscuit ext info extension_name

    Args:
        name (str): The name of the extension"""

    def f(app: App, name=name) -> None:
        data = app.extensions_manager.find_extension_by_name(name)
        if data:
            click.echo(f"Name: {name}")
            click.echo(f"Author: {data[1]}")
            click.echo(f"Description: {data[2]}")
            # TODO: click.echo(f"Version: {data[3]}")
        else:
            click.echo(f"Could not find extension {name}")

    return f


@ext.command()
@click.argument("name")
def install(name: str) -> typing.Callable[[App], None]:
    """Install an extension by name

    Example:
        biscuit ext install extension_name

    Args:
        name (str): The name of the extension
    """

    def f(app: App, name=name) -> None:
        if app.extensions_manager.install_extension_from_name(name):
            click.echo(f"Installed extension {name} successfully")
        else:
            click.echo(f"Could not find extension {name}")

    return f


@ext.command()
@click.argument("name")
def uninstall(name: str) -> typing.Callable[[App], None]:
    """Uninstall an extension by name

    Example:
        biscuit ext uninstall extension_name

    Args:
        name (str): The name of the extension
    """

    def f(app: App, name=name) -> None:
        if app.extensions_manager.uninstall_extension_from_name(name):
            click.echo(f"Uninstalled extension {name} successfully")
        else:
            click.echo(f"Could not find extension {name}")

    return f


@ext.command()
def create():
    """Create a new extension from template

    This command will create a new extension from the template

    NOTE: This command is not yet implemented
    """

    click.echo("Extension created!")


@ext.command()
def test():
    """Test the extension

    This command will test your custom extension

    NOTE: This command is not yet implemented
    """

    click.echo("Extension tested!")


def register(cli: click.Group):
    cli.add_command(ext)



from __future__ import annotations

import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import IconLabelButton


class SubMenuItem(IconLabelButton):
    """SubMenu"""

    def __init__(self, main, master, text, *args, **kwargs) -> None:
        """Create a submenu item

        Args:
            master: The parent widget
            text: The text to display on the menu item
        """

        super().__init__(
            master,
            text,
            Icons.CHEVRON_RIGHT,
            expandicon=False,
            iconsize=10,
            iconside=tk.RIGHT,
            *args,
            **kwargs,
        )
        self.main = main
        self.text_label.config(padx=14)

        from .menu import Menu

        class SubMenu(Menu):
            submenu = True

            def __init__(self, master, main: Menu, *args, **kwargs) -> None:
                super().__init__(master, *args, **kwargs)
                self.main = main

            def get_coords(self) -> None:
                return (
                    self.master.winfo_rootx() + self.master.winfo_width(),
                    self.master.winfo_rooty(),
                )

            def event_chosen(self) -> None:
                super().event_chosen()
                self.main.event_chosen()

        self.menu = SubMenu(self, main, text)

        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

    def on_enter(self, e) -> None:
        self.main.hold_focus = True
        self.menu.show()
        super().on_enter(e)

    def on_leave(self, e=None) -> None:
        self.main.hold_focus = False

        # TODO: Keeping highlight on the submenuitem
        # following won't work because hide() is not called (somehow magically happens)

        # if self.menu.active:
        #     return

        super().on_leave(e)
        # self.menu.hide()

    def on_click(self, *_) -> None: ...



import json

import requests


class Releases:
    """A class to interact with the GitHub API to get release information.
    If extensions require any prebuilt binaries, use the methods in this class."""

    def __init__(self, base=None):
        self.base = base
        
    def get(self, url: str):
        resp = requests.get(url)
        if resp.status_code == 200:
            return json.loads(resp.text)
    
    def get_latest_release(self, owner: str, repo: str):
        return self.get(f"https://api.github.com/repos/{owner}/{repo}/releases/latest")
    
    def get_release(self, owner: str, repo: str, tag: str):
        return self.get(f"https://api.github.com/repos/{owner}/{repo}/releases/tags/{tag}")
    
    def get_latest_windows_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp['assets']:
            if "-windows-" in asset['name'].lower():
                return asset['browser_download_url']
            
    def get_latest_macos_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp['assets']:
            if "-mac-" in asset['name'].lower():
                return asset['browser_download_url']

    def get_latest_linux_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp['assets']:
            if "-linux-" in asset['name'].lower():
                return asset['browser_download_url']

    def get_windows_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp['assets']:
            if "-windows-" in asset['name'].lower():
                return asset['browser_download_url']

    def get_macos_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp['assets']:
            if "-mac-" in asset['name'].lower():
                return asset['browser_download_url']

    def get_linux_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp['assets']:
            if "-linux-" in asset['name'].lower():
                return asset['browser_download_url']



from __future__ import annotations

import sqlite3


class FixedSizeStack:
    """
    A stack with a fixed size. If the stack is full, the oldest item is removed
    when a new item is added.
    """

    def __init__(self, master, name: str, capacity=5):
        self.base = master.base
        self.name = name
        self.capacity = capacity
        self.stack = []

    def __iter__(self):
        return iter(self.stack)

    @property
    def list(self):
        return [(i, lambda _, i=i: self.base.open(i)) for i in self.stack[::-1]]

    def push(self, item):
        if len(self.stack) == self.capacity:
            # Remove the item at the bottom (oldest item) if the stack is full
            self.stack.pop(0)
        if item in self.stack:
            # Remove the item if it already exists
            self.stack.remove(item)
        self.stack.append(item)

    def pop(self):
        if not self.is_empty():
            return self.stack.pop()

    def is_empty(self):
        return not len(self.stack)

    def __len__(self):
        return len(self.stack)

    def clear(self):
        self.stack.clear()

    def dump_sqlite(self, cursor: sqlite3.Cursor) -> None:
        """Dump the stack to the database.

        Args:
            cursor (sqlite3.Cursor): the cursor to the database"""

        # print(self.name, self.stack)
        cursor.execute(f"DELETE FROM {self.name};")
        cursor.executemany(
            f"INSERT INTO {self.name} (path) VALUES (?);",
            [(item,) for item in self.stack],
        )

    def load_sqlite(self, cursor: sqlite3.Cursor) -> FixedSizeStack:
        """Load the stack from the database.

        Args:
            cursor (sqlite3.Cursor): the cursor to the database"""

        cursor.execute(f"SELECT path FROM {self.name};")
        self.stack = [item[0] for item in cursor.fetchall()]
        return self



from ..icons import Icons
from .labels import Label
from .native import Menubutton


class Icon(Label):
    """Simple label using codicons"""

    def __init__(
        self, master, icon: Icons = "", iconsize: int = 12, *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(font=("codicon", iconsize))
        self.set_icon(icon)

    def set_icon(self, icon: str) -> None:
        self.config(text=icon)

    def set_color(self, color: str) -> None:
        self.config(fg=color)


class IconButton(Menubutton):
    """Icon button using codicons"""

    def __init__(
        self,
        master,
        icon: Icons,
        event=lambda *_: ...,
        icon2: Icons = "",
        iconsize=12,
        highlighted=False,
        hfg_only=False,
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(
            cursor="hand2",
            **(
                (
                    self.base.theme.utils.iconbutton_hfg
                    if hfg_only
                    else self.base.theme.utils.iconbutton
                )
                if not highlighted
                else self.base.theme.utils.button
            )
        )
        self.icons = [icon, icon2]
        self.icon2 = icon2
        self.switch = False

        self.event = event
        self.config(text=icon, font=("codicon", iconsize))

        self.bind("<Button-1>", self.onclick)

    def set_callback(self, event) -> None:
        self.event = event

    def onclick(self, *args) -> None:
        try:
            try:
                self.event(*args)
            except:
                self.event()

            self.v_onclick()
            self.toggle_icon()
        except Exception as e:
            print(e)

    def v_onclick(self) -> None: ...

    def set_icon(self, icon) -> None:
        self.config(text=icon)

    def toggle_icon(self) -> None:
        if not self.icon2:
            return

        self.switch = not self.switch
        self.config(text=self.icons[self.switch])

    def reset_icon(self) -> None:
        self.switch = False
        self.config(text=self.icons[self.switch])



import tkinter as tk

from ..icons import Icons
from .icon import IconButton
from .native import Frame


class Closable(Frame):
    """For implementing a list of closeable items.
    A varient of IconLabelButton that comes with a close icon
    on the right side of the button.
    """

    def __init__(
        self,
        master,
        text=None,
        icon: Icons = "",
        callback=lambda *_: None,
        closefn=lambda *_: None,
        iconside=tk.LEFT,
        padx=1,
        pady=1,
        fg=None,
        bg=None,
        hfg=None,
        hbg=None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, padx=padx, pady=pady, *args, **kwargs)
        self.callback = callback
        self.closefn = closefn

        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.utils.iconlabelbutton.values()
        )
        if fg:
            self.fg = fg
        if bg:
            self.bg = bg
        if hfg:
            self.hfg = hfg
        if hbg:
            self.hbg = hbg

        self.config(bg=self.bg)
        self.text = text
        self.icon = icon

        self.icon_label = None
        self.text_label = None

        if icon:
            self.icon_label = tk.Label(
                self,
                text=self.icon,
                anchor=tk.CENTER,
                bg=self.bg,
                fg=self.fg,
                font=("codicon", 14),
            )
            self.icon_label.pack(side=iconside, fill=tk.BOTH, padx=(10, 0))

        if text:
            self.text_label = tk.Label(
                self,
                text=self.text,
                anchor=tk.CENTER,
                pady=2,
                bg=self.bg,
                fg=self.fg,
                font=self.base.settings.uifont,
            )
            self.text_label.pack(
                side=iconside, fill=tk.BOTH, expand=True, padx=10 if not icon else 0
            )

        self.close_btn = IconButton(self, Icons.CLOSE, self.closefn)
        self.close_btn.config(bg=self.bg, fg=self.fg)
        self.close_btn.pack(side=tk.RIGHT, fill=tk.BOTH)

        self.config_bindings()
        self.visible = False

    def config_bindings(self) -> None:
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

        self.bind("<Button-1>", self.on_click)
        if self.text:
            self.text_label.bind("<Button-1>", self.on_click)
        if self.icon:
            self.icon_label.bind("<Button-1>", self.on_click)

    def on_enter(self, *_) -> None:
        self.config(bg=self.hbg)
        if self.text:
            self.text_label.config(bg=self.hbg, fg=self.hfg)
        if self.icon:
            self.icon_label.config(bg=self.hbg, fg=self.hfg)
        self.close_btn.config(bg=self.hbg, fg=self.hfg)

    def on_leave(self, *_) -> None:
        self.config(bg=self.bg)
        if self.text:
            self.text_label.config(bg=self.bg, fg=self.fg)
        if self.icon:
            self.icon_label.config(bg=self.bg, fg=self.fg)
        self.close_btn.config(bg=self.bg, fg=self.fg)

    def on_click(self, *_) -> None:
        self.callback()

    def change_text(self, text) -> None:
        """Change the text of the item"""

        self.text_label.config(text=text)

    def change_icon(self, icon) -> None:
        """Change the icon of the item"""

        self.icon_label.config(text=icon)

    def set_pack_data(self, **kwargs) -> None:
        self.pack_data = kwargs

    def get_pack_data(self):
        return self.pack_data

    def show(self) -> None:
        """Show the item"""

        if not self.visible:
            self.visible = True
            self.pack(**self.get_pack_data())

    def hide(self) -> None:
        """Hide the item"""

        if self.visible:
            self.visible = False
            self.pack_forget()



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame

from .editors import EditorsManager
from .panel import Panel

if typing.TYPE_CHECKING:
    ...


class Content(Frame):
    """Content Pane

    - Contains the EditorsPane and Panel
    - Manages the visibility of the Panel and the EditorPane
    """

    def __init__(self, master: Frame, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.editorspane = EditorsManager(self)
        self.panel = Panel(self)
        self._panel_enabled = False
        self._panel_maxed = False

        self.editorspane.pack(fill=tk.BOTH, expand=True)

    def show_panel(self) -> None:
        if not self._panel_enabled:
            self.toggle_panel()

    def toggle_panel(self, *_) -> None:
        if self._panel_enabled:
            self.panel.pack_forget()
            self.editorspane.pack(fill=tk.BOTH, expand=True)
        else:
            if self._panel_maxed:
                self.panel.pack(fill=tk.BOTH, pady=(1, 0), expand=True)
                self.editorspane.pack_forget()
            else:
                self.panel.pack(fill=tk.BOTH, pady=(1, 0))

            if not self.panel.terminals.active_terminal:
                self.panel.terminals.open_terminal()

        self._panel_enabled = not self._panel_enabled

    def toggle_max_panel(self, *_) -> None:
        if self._panel_maxed:
            self.panel.pack_forget()
            self.editorspane.pack(fill=tk.BOTH, expand=True)
            self.panel.pack(fill=tk.BOTH, pady=(1, 0))
        else:
            self.editorspane.pack_forget()
            self.panel.pack(fill=tk.BOTH, pady=(1, 0), expand=True)

        self._panel_maxed = not self._panel_maxed

    def pack(self):
        super().pack(side=tk.LEFT, fill=tk.BOTH, expand=True)



from __future__ import annotations

import os
import sqlite3
import typing

from .common import ActionSet, FixedSizeStack

if typing.TYPE_CHECKING:
    from . import App


class HistoryManager:
    """Manages the history of opened files and folders.

    Manages the history of opened files and folders.
    Uses an sqlite3 database to store the history.
    """

    def __init__(self, base: App) -> None:
        self.base = base
        self.path = os.path.join(self.base.datadir, "history.db")

        self.db = sqlite3.connect(self.path)
        self.cursor = self.db.cursor()

        self.cursor.executescript(
            """
            CREATE TABLE IF NOT EXISTS file_history (path TEXT NOT NULL);
            CREATE TABLE IF NOT EXISTS folder_history (path TEXT NOT NULL);
            """
        )

        self.file_history = FixedSizeStack(self, "file_history").load_sqlite(
            self.cursor
        )
        self.folder_history = FixedSizeStack(self, "folder_history").load_sqlite(
            self.cursor
        )

    def generate_actionsets(self) -> None:
        self.base.palette.register_actionset(
            lambda: ActionSet("Recent files", "recentf:", self.file_history.list)
        )
        self.base.palette.register_actionset(
            lambda: ActionSet("Recent folders", "recentd:", self.folder_history.list)
        )

    def register_file_history(self, path: str) -> None:
        self.file_history.push(path)

    def register_folder_history(self, path: str) -> None:
        self.folder_history.push(path)

    def dump(self) -> None:
        self.file_history.dump_sqlite(self.cursor)
        self.folder_history.dump_sqlite(self.cursor)
        self.db.commit()

    def clear_history(self) -> None:
        self.file_history.clear()
        self.folder_history.clear()



import sv_ttk

from .theme import *


class CatppuccinMocha(Theme):
    name = "Catppuccin Mocha"

    biscuit = "#f5e0dc"  # Original biscuit color from the Catppuccin Mocha palette
    biscuit_dark = "#cba6f7"  # A darker accent color closer to the theme

    border = "#313244"  # Darker border color to match the overall theme
    disabled = "#585b70"  # Correct shade for disabled elements

    primary_background = "#1e1e2e"  # This is correct (mantle in Catppuccin Mocha)
    primary_foreground = "#cdd6f4"  # This is correct (text color)
    primary_background_highlight = (
        "#313244"  # Slightly darker for highlighting background
    )
    primary_foreground_highlight = "#f5e0dc"  # Lighter highlight for readability

    secondary_background = "#181825"  # Slightly darker than primary background
    secondary_foreground = "#cdd6f4"  # This is correct
    secondary_background_highlight = "#313244"  # Matches the highlight from primary
    secondary_foreground_highlight = "#f5e0dc"  # Same as primary foreground highlight

    # Syntax highlighting
    Keyword = "#f5e0dc"  # Correct light pink for keywords
    Name = "#89b4fa"  # Correct blue for names
    Function = "#f9e2af"  # Correct yellow for function names
    String = "#a6e3a1"  # Green color for strings
    Number = "#fab387"
    Comment = {
        "foreground": "#6c7086",
        "font": {"slant": "italic"},
    }  # Gray for comments with italic style
    Punctuation = "#f5e0dc"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_dark_theme()

        self.editors.linenumbers.number.foreground = "#cdd6f4"
        self.editors.linenumbers.number.highlightforeground = "#cdd6f4"

        self.editors.linenumbers.breakpoint.background = "#f38ba8"
        self.editors.linenumbers.breakpoint.highlightbackground = "#f38ba8"



import tkinter as tk

from git import Optional
from hintedtext import HintedEntry

from ..icons import Icons
from .icon import IconButton
from .native import Frame


class ButtonsEntry(Frame):
    """Entry containing icon buttons on the right

    Args:
        master: Parent widget
        hint: Entry hint
        buttons: List of tuples containing icon, event and optional second icon (toggled)
        textvariable: Entry text variable
    """

    def __init__(
        self,
        master,
        hint: str,
        buttons: list[tuple[Icons, callable, Optional[str]]] = [],
        textvariable=None,
        *args,
        **kwargs
    ):
        """Create a new ButtonsEntry

        Args:
            master: Parent widget
            hint: Entry hint
            buttons: List of tuples containing button icon, event and optional second icon (toggled)
            textvariable: Entry text variable"""

        super().__init__(master, *args, **kwargs)
        self.config(padx=1, pady=1, bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.entry = HintedEntry(
            self,
            relief=tk.FLAT,
            font=self.base.settings.uifont,
            bd=5,
            hint=hint,
            **self.base.theme.utils.buttonsentry,
            textvariable=textvariable
        )
        self.entry.grid(row=0, column=0, sticky=tk.NSEW)

        self.column = 1
        self.add_buttons(buttons)

    def add_button(self, icon: Icons, event=lambda _: None, icon2: str = None):
        """Add a button to the entry"""

        b = IconButton(self, icon, event, icon2)
        b.grid(row=0, column=self.column, sticky="")
        b.config(**self.base.theme.utils.buttonsentry.button)
        self.column += 1

    def add_buttons(self, buttons):
        """Add multiple buttons to the entry"""

        for btn in buttons:
            self.add_button(*btn)

    def get(self, *args):
        """Get the entry text"""

        return self.entry.get(*args)

    def clear(self):
        """Clear the entry text"""

        return self.entry.delete(0, tk.END)

    def delete(self, *args):
        """Delete the entry text"""

        return self.entry.delete(*args)

    def insert(self, *args):
        """Insert text into the entry"""

        return self.entry.insert(*args)



# Palette: https://github.com/morhetz/gruvbox

import sv_ttk

from .theme import *


class GruvboxLight(Theme):
    name = "gruvbox light"

    biscuit = "#8fbf7c"
    biscuit_dark = "#3c3836"

    border = "#3c3836"
    disabled = "#737373"

    primary_background = "#f9f5d7"
    primary_foreground = "#ebdbb2"
    primary_background_highlight = "#282828"
    primary_foreground_highlight = "#ebdbb2"

    secondary_background = "#282828"
    secondary_foreground = "#ebdbb2"
    secondary_background_highlight = "#1d2021"
    secondary_foreground_highlight = "#ebdbb2"

    # syntax
    Keyword = "#f84b3c"
    Name = "#ebdab4"
    Function = "#8fbf7f"
    String = "#b8ba37"
    Number = "#d28797"
    Comment = {"foreground": "#928375", "font": {"slant": "italic"}}
    Punctuation = "#8fbf7f"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_dark_theme()

        self.editors.linenumbers.number.foreground = "#928375"
        self.editors.linenumbers.number.highlightforeground = "#928375"

        self.editors.linenumbers.breakpoint.background = "#df463a"
        self.editors.linenumbers.breakpoint.highlightbackground = "#df463a"



from __future__ import annotations

import tkinter as tk
import typing

import mistune
from mistune import InlineParser
from mistune.plugins.abbr import abbr
from mistune.plugins.def_list import def_list
from mistune.plugins.footnotes import footnotes
from mistune.plugins.formatting import strikethrough
from mistune.plugins.speedup import speedup
from mistune.plugins.table import table
from mistune.plugins.task_lists import task_lists
from mistune.plugins.url import url
from pygments import highlight
from pygments.formatters import html
from pygments.lexers import get_lexer_by_name
from pygments.style import Style
from tkinterweb import HtmlFrame

from biscuit.common.ui import Frame, Scrollbar

if typing.TYPE_CHECKING:
    from ...settings.config import Theme


class HighlightRenderer(mistune.HTMLRenderer):
    def __init__(self, style, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.style = style
        self.formatter = html.HtmlFormatter(style=self.style)

    def block_code(self, code, info=None):
        try:
            if info:
                lexer = get_lexer_by_name(info, stripall=True)
                return highlight(code, lexer, self.formatter)
        except Exception as e:
            print(e)

        return "<pre><code>" + mistune.escape(code) + "</code></pre>"


class BiscuitStyle(Style):
    def __init__(self, theme: Theme):
        super().__init__()
        self.theme = theme
        self.styles = self.theme.syntax

    def __iter__(self):
        for token, color in self.theme.syntax.items():
            if isinstance(color, dict):
                yield token, {
                    "color": color["foreground"][1:],
                    "bold": False,
                    "italic": False,
                    "underline": False,
                    "bgcolor": None,
                    "border": None,
                }
            else:
                yield token, {
                    "color": color[1:],
                    "bold": False,
                    "italic": False,
                    "underline": False,
                    "bgcolor": None,
                    "border": None,
                }


class Renderer(Frame):
    """Renderer for the AI assistant chat view.

    The Renderer is used to render the chat messages in the AI chat view.
    - The Renderer uses the HtmlFrame widget to display the chat messages.
    - The chat messages support markdown formatting."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.style = BiscuitStyle(self.base.theme)

        self.renderer = HighlightRenderer(self.style, escape=False)
        self.formatter = self.renderer.formatter
        self.markdown = mistune.Markdown(
            renderer=self.renderer,
            inline=InlineParser(),
            plugins=[
                abbr,
                def_list,
                footnotes,
                strikethrough,
                speedup,
                table,
                task_lists,
                url,
            ],
        )
        self.htmlframe = HtmlFrame(
            self, messages_enabled=False, vertical_scrollbar=False
        )
        self.scrollbar = Scrollbar(
            self,
            orient=tk.VERTICAL,
            command=self.htmlframe.yview,
            style="EditorScrollbar",
        )

        self.sparkles = f"<h4 color={self.base.theme.biscuit}>✨ Bikkis</h4> "

        self.htmlframe.html.config(yscrollcommand=self.scrollbar.set)
        self.htmlframe.html.shrink(True)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.htmlframe.grid(row=0, column=1, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, column=3, sticky=tk.NS)

        self.header = "<html><head></head><body>"
        self.footer = "</body></html>"
        self.content = ""

        t = self.base.theme
        t = self.base.theme
        pygments_css = self.formatter.get_style_defs(".highlight")
        self.css = f"""
            {pygments_css}
            CODE, PRE {{
                font-family: {self.base.settings.font['family']};
                font-size: {self.base.settings.font['size']}pt;
                background-color: {t.border};
                padding: 2px;
            }}
            BODY {{
                background-color: {t.primary_background};
                color: {t.primary_foreground};
                font-family: {self.base.settings.uifont['family']};
                font-size: {self.base.settings.uifont['size']}pt;
            }}
            img {{
                max-width: 100%;
                height: auto;
            }}

            hr {{
                border: 0;
                border-top: 1px solid {t.border};
                max-width: 100%;
            }}
            li{{
                margin-left:1px;
            }}

            tr, th, td {{
                border: 1px solid {t.border};
            }}

            :link    {{ color: {t.biscuit}; }}
            :visited {{ color: {t.biscuit_dark}; }}
            INPUT, TEXTAREA, SELECT, BUTTON {{ 
                background-color: {t.secondary_background};
                color: {t.secondary_foreground_highlight};
            }}
            INPUT[type="submit"],INPUT[type="button"], INPUT[type="reset"], BUTTON {{
                background-color: {t.primary_background};
                color: {t.primary_foreground};
                color: tcl(::tkhtml::if_disabled {t.primary_background}{t.primary_foreground_highlight});
            }}
            """

    def write(self, content: str, sparkles: bool = False) -> None:
        if sparkles:
            self.content += self.sparkles
        self.content += self.markdown(content)
        self.htmlframe.load_html(self.content)
        self.htmlframe.add_css(self.css)



import os
import tkinter as tk
from typing import Callable, List

from biscuit.common.icons import Icons
from biscuit.common.ui import Icon, Shortcut
from biscuit.common.ui.native import Frame, Label


class RecentItem(Frame):
    """Recent Menu Item"""

    def __init__(self, master, path: str, callback: Callable, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.text = path
        self.callback = callback

        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.editors.values()

        self.label = Label(
            self,
            text=os.path.basename(path),
            fg=self.base.theme.biscuit_light,
        )
        self.label.bind("<Button-1>", self.callback)
        self.label.pack(side=tk.LEFT, padx=5)

        self.fullpath_label = Label(
            self, text=os.path.abspath(path), fg=self.base.theme.biscuit
        )
        self.fullpath_label.pack(side=tk.RIGHT, padx=5)

        # self.close_button = Icon(
        #     self, icon=Icons.CLOSE, fg=self.base.theme.biscuit_light
        # )
        # self.close_button.pack(side=tk.LEFT, padx=5)

        self.config(padx=5, pady=5)
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)
        self.bind("<Button-1>", self.callback)
        self.label.bind("<Button-1>", self.callback)
        self.fullpath_label.bind("<Button-1>", self.callback)
        self.on_leave()

    def on_enter(self, *_) -> None:
        self.config(bg=self.hbg)
        self.label.config(bg=self.hbg)
        self.fullpath_label.config(bg=self.hbg)
        # self.close_button.config(bg=self.base.theme.border)

    def on_leave(self, *_) -> None:
        self.config(bg=self.bg)
        self.label.config(bg=self.bg)
        self.fullpath_label.config(bg=self.bg)

        # self.close_button.config(bg=self.bg)


class QuickItem(Frame):
    """Quick Menu Item"""

    def __init__(
        self,
        master,
        text: str,
        icon: str,
        callback: Callable,
        shortcut: List[str],
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.text = text
        self.icon = icon
        self.callback = callback
        self.shortcut = shortcut

        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.editors.values()

        self.icon = Icon(self, icon=icon, fg=self.base.theme.biscuit_light)
        self.icon.pack(side=tk.LEFT, padx=5)

        self.label = Label(self, text=text, fg=self.base.theme.biscuit_light)
        self.label.bind("<Button-1>", self.callback)
        self.label.pack(side=tk.LEFT, padx=5)

        self.shortcutlabel = Shortcut(self, shortcut)
        self.shortcutlabel.pack(side=tk.RIGHT, padx=5)
        self.shortcutlabel.bind("<Button-1>", self.callback)

        self.config(padx=5, pady=5)
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)
        self.bind("<Button-1>", self.callback)
        self.on_leave()

    def on_enter(self, *_) -> None:
        self.config(bg=self.base.theme.border)
        self.icon.config(bg=self.base.theme.border)
        self.label.config(bg=self.base.theme.border)
        self.shortcutlabel.config(bg=self.base.theme.border)

    def on_leave(self, *_) -> None:
        self.config(bg=self.bg)
        self.icon.config(bg=self.bg)
        self.label.config(bg=self.bg)
        self.shortcutlabel.config(bg=self.bg)



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common import Icons
from biscuit.common.ui import Bubble, Menubutton

if typing.TYPE_CHECKING:
    from biscuit.views import SideBarView

    from .activitybar import ActivityBar


class SBubble(Bubble):
    def get_pos(self) -> str:
        return (
            f"+{int(self.master.winfo_rootx() + (self.master.winfo_width() - self.winfo_width())/2)}"
            + f"+{self.master.winfo_rooty() - self.master.winfo_height() - 10}"
        )


class ActionButton(Menubutton):
    """Action buttons for activity bar

    Action buttons are used to switch between views in the sidebar,
    view instances are attached to these buttons."""

    def __init__(self, master: ActivityBar, view: SideBarView, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        """Initializes the action button.
        
        Args:
            master (ActivityBar): The activity bar.
            view (SidebarView): The view to attach to the button."""
        self.master: ActivityBar = master

        self.view = view
        self.enabled = False

        self.bubble = SBubble(self, text=view.name)
        self.bind("<Enter>", self.bubble.show)
        self.bind("<Leave>", self.bubble.hide)

        self.config(
            text=view.__icon__,
            relief=tk.FLAT,
            font=("codicon", 12),
            cursor="hand2",
            padx=5,
            pady=5,
            **self.base.theme.layout.sidebar.actionbar.slot,
        )
        self.pack(fill=tk.X, side=tk.TOP)

        self.bind("<Button-1>", self.toggle)

    def toggle(self, *_) -> None:
        if not self.enabled:
            self.master.set_active_slot(self)
            self.enable()
        else:
            self.disable()

        self.bubble.hide()

    def enable(self) -> None:
        if not self.enabled:
            self.master.sidebar.pack()
            self.view.grid(column=0, row=1, sticky=tk.NSEW)
            self.config(
                fg=self.base.theme.layout.sidebar.actionbar.slot.selectedforeground
            )
            self.enabled = True

    def disable(self) -> None:
        if self.enabled:
            self.master.sidebar.hide()
            self.view.grid_remove()
            self.config(fg=self.base.theme.layout.sidebar.actionbar.slot.foreground)
            self.enabled = False



import os

from biscuit.common.icons import Icons
from biscuit.common.ui import Label

from ..terminalbase import TerminalBase


class CommandPrompt(TerminalBase):
    """
    Command Prompt - Checks for Command Prompt executable in path and opens that in terminal.
    Shows Not Available in case variable is not set.
    """

    shell = "C:\\windows\\system32\\cmd.exe"
    name = "Default"
    icon = Icons.TERMINAL_CMD

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        if not self.shell:
            Label(
                self, text="Command Prompt not available, report an issue otherwise."
            ).grid()
            self.name = "Not Available"
            self.icon = "error"
            return

        self.start_service()



from __future__ import annotations

import typing

from biscuit.common import Menu

if typing.TYPE_CHECKING:
    from .directorytree import DirectoryTree


class ExplorerMenu(Menu):
    def get_coords(self, e) -> list:
        return e.widget.winfo_rootx(), e.widget.winfo_rooty() + e.widget.winfo_height()


class DirectoryContextMenu(Menu):
    def __init__(self, master: DirectoryTree, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master

        self.add_command("New file...", lambda: self.base.palette.show("newfile:"))
        self.add_command("New Folder...", lambda: self.base.palette.show("newfolder:"))
        self.add_command("Reveal in File Explorer", self.master.reveal_in_explorer)
        self.add_command("Open in Integrated Terminal", self.master.open_in_terminal)
        self.add_command("Reopen editor", self.master.reopen_editor)
        # self.add_separator()
        # self.add_command("Copy", self.master.new_file)
        # self.add_command("Cut", self.master.new_file)
        # self.add_command("Paste", self.master.new_file)
        self.add_separator()
        self.add_command("Copy Path", self.master.copy_path)
        self.add_command("Copy Relative Path", self.master.copy_relpath)
        self.add_command("Copy Name", self.master.copy_name)
        self.add_command(
            "Copy Name without Extension", self.master.copy_name_without_extension
        )
        self.add_separator()
        self.add_command("Attach to bikkis...", self.master.attach_to_chat)
        self.add_command("Add to .gitignore", self.master.add_to_gitignore)
        self.add_command("Exclude from .gitignore", self.master.exclude_from_gitignore)
        self.add_separator()
        self.add_command("Rename...", lambda: self.base.palette.show("rename:"))
        self.add_command("Delete", self.master.delete_item)

    def get_coords(self, e) -> list:
        return e.x_root, e.y_root

    def show(self, *e) -> None:
        super().show(*e)



import sv_ttk

from .theme import *


class Dark(Theme):
    name = "Myst"

    biscuit = "#EE720D"
    biscuit_dark = "#EE720D"

    border = "#141414"
    disabled = "#141414"

    primary_background = "#141414"
    primary_foreground = "#757575"
    primary_background_highlight = "#1C1C1C"
    primary_foreground_highlight = "#757575"

    secondary_background = "#222222"
    secondary_foreground = "#ffdadb"
    secondary_background_highlight = "#2A2A2A"
    secondary_foreground_highlight = "#ffdadb"

    # syntax
    Keyword = "#f84b3c"
    Name = "#ebdab4"
    Function = "#8fbf7f"
    String = "#b8ba37"
    Number = "#d28797"
    Comment = {"foreground": "#928375", "font": {"slant": "italic"}}
    Punctuation = "#8fbf7f"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_dark_theme()

        self.editors.linenumbers.number.foreground = "#928375"
        self.editors.linenumbers.number.highlightforeground = "#928375"

        self.editors.linenumbers.breakpoint.background = "#df463a"
        self.editors.linenumbers.breakpoint.highlightbackground = "#df463a"



from __future__ import annotations

import os
import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton

if typing.TYPE_CHECKING:
    from .breadcrumbs import BreadCrumbs


class HistoryNavigation(Frame):
    """History navigation widget

    History navigation widget is used to navigate the history of the cursor
    position in the editor.
    """

    def __init__(self, master: BreadCrumbs, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: BreadCrumbs = master
        self.config(bg=self.base.theme.border)

        self.left = IconButton(
            self,
            Icons.ARROW_LEFT,
            event=self.go_down,
            iconsize=12,
            padx=7,
            hfg_only=True,
        )
        self.left.config(**self.base.theme.editors.breadcrumbs.item)

        self.right = IconButton(
            self,
            Icons.ARROW_RIGHT,
            event=self.go_up,
            iconsize=12,
            padx=7,
            hfg_only=True,
        )
        self.right.config(**self.base.theme.editors.breadcrumbs.item)

        self.left.config(state=tk.DISABLED)
        self.right.config(state=tk.DISABLED)

        self.left.pack(side=tk.LEFT, fill=tk.BOTH)
        self.right.pack(side=tk.RIGHT, fill=tk.BOTH)

    # TODO: Implement go_up and go_down methods
    def go_up(self) -> None: ...
    def go_down(self) -> None: ...



import tkinter as tk

from biscuit.common import Menu


class TerminalMenu(Menu):
    def get_coords(self, e: tk.Event):
        return (
            e.widget.winfo_rootx() + e.widget.winfo_width() - self.winfo_width(),
            e.widget.winfo_rooty() + e.widget.winfo_height(),
        )



import tkinter as tk

from ..icons import Icons
from .native import Frame, Menubutton


class Button(Menubutton):
    """A flat style button"""

    def __init__(self, master, text, command=lambda _: None, *args, **kwargs) -> None:
        super().__init__(master, text=text, *args, **kwargs)
        self.config(
            pady=5,
            font=self.base.settings.uifont,
            cursor="hand2",
            **self.base.theme.utils.button
        )
        self.set_command(command)

    def set_command(self, command) -> None:
        """Set the command for the button"""

        self.bind("<Button-1>", command)


class IconLabelButton(Frame):
    """Icon label button with both text and icon

    Args:
        text (str): Text to display on the button
        icon (str): Icon to display on the button
        callback (function): Function to call when the button is clicked
        iconside (str): Side to display the icon
        expandicon (bool): Expand the icon
        highlighted (bool): Highlight the button
        iconsize (int): Size of the icon
        icon_visible (bool): Initial state of the icon visibility"""

    def __init__(
        self,
        master,
        text=None,
        icon: Icons = "",
        callback=lambda *_: None,
        iconside=tk.LEFT,
        expandicon=True,
        highlighted=False,
        iconsize=14,
        icon_visible=True,
        padx=5,
        pady=1,
        bg="",
        hbg="",
        hfg_only=False,
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, padx=padx, pady=pady, *args, **kwargs)

        self.bg, self.fg, self.hbg, self.hfg = (
            (
                self.base.theme.utils.iconlabelbutton_hfg.values()
                if hfg_only
                else self.base.theme.utils.iconlabelbutton.values()
            )
            if not highlighted
            else self.base.theme.utils.button.values()
        )
        if bg:
            self.bg = bg

        if hbg:
            self.hbg = hbg

        self.config(bg=self.bg)
        self.text = text
        self.icon = icon
        self.icon_visible = icon_visible
        self.callback = callback
        self.codicon = self.icon

        if icon:
            self.icon_label = tk.Label(
                self,
                text=self.codicon if self.icon_visible else "    ",
                anchor=tk.E,
                bg=self.bg,
                fg=self.fg,
                font=("codicon", iconsize),
                cursor="hand2",
            )
            self.icon_label.pack(side=iconside, fill=tk.BOTH, expand=expandicon)

        if text:
            self.text_label = tk.Label(
                self,
                text=self.text,
                anchor=tk.W,
                pady=2,
                bg=self.bg,
                fg=self.fg,
                font=self.base.settings.uifont,
                cursor="hand2",
            )
            self.text_label.pack(side=iconside, fill=tk.BOTH, expand=True)

        self.config_bindings()
        self.visible = False

    def config_bindings(self) -> None:
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

        self.bind("<Button-1>", self.on_click)
        if self.text:
            self.text_label.bind("<Button-1>", self.on_click)
        if self.icon:
            self.icon_label.bind("<Button-1>", self.on_click)

    def on_enter(self, *_) -> None:
        self.config(bg=self.hbg)
        if self.text:
            self.text_label.config(bg=self.hbg, fg=self.hfg)
        if self.icon:
            self.icon_label.config(bg=self.hbg, fg=self.hfg)

    def on_leave(self, *_) -> None:
        self.config(bg=self.bg)
        if self.text:
            self.text_label.config(bg=self.bg, fg=self.fg)
        if self.icon:
            self.icon_label.config(bg=self.bg, fg=self.fg)

    def on_click(self, *_) -> None:
        self.callback()

    def toggle_icon(self) -> None:
        try:
            self.icon_label.config(
                text=self.codicon if not self.icon_visible else "    "
            )
            self.icon_visible = not self.icon_visible
        except Exception:
            pass

    def change_text(self, text) -> None:
        try:
            self.text_label.config(text=text)
        except Exception:
            pass

    def change_icon(self, icon) -> None:
        try:
            self.icon_label.config(text=icon)
        except Exception:
            pass

    def set_pack_data(self, **kwargs) -> None:
        self.pack_data = kwargs

    def get_pack_data(self):
        return self.pack_data

    def show(self) -> None:
        if not self.visible:
            self.visible = True
            self.pack(**self.get_pack_data())

    def hide(self) -> None:
        if self.visible:
            self.visible = False
            self.pack_forget()



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconLabelButton, WrappingLabel


class ChangesTreePlaceholder(Frame):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)
        self.columnconfigure(0, weight=1)

        WrappingLabel(
            self,
            font=self.base.settings.uifont,
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
            text="Open a folder containing a git repository."
        ).grid(row=0, sticky=tk.EW)

        # TODO add init repo button if no repo is found for opened folder

    #     open_btn = IconLabelButton(
    #         self,
    #         text="Open Folder",
    #         icon=Icons.FOLDER,
    #         callback=self.open_folder,
    #         pady=2,
    #         highlighted=True,
    #     )
    #     open_btn.grid(row=1, pady=5, sticky=tk.EW)

    #     clone_btn = IconLabelButton(
    #         self,
    #         text="Clone Repository",
    #         icon=Icons.GIT_COMMIT,
    #         callback=self.clone_repo,
    #         pady=2,
    #         highlighted=True,
    #     )
    #     clone_btn.grid(row=2, pady=5, sticky=tk.EW)

    # def open_folder(self, *_) -> None:
    #     self.base.commands.open_directory()

    # def clone_repo(self, *_) -> None:
    #     self.base.palette.show("clone:")



import os
import pathlib
import platform
import shutil
import subprocess
import threading
import tkinter as tk
from tkinter.messagebox import askyesno
from typing import Iterator

import pyperclip

from biscuit.common.icons import Icons
from biscuit.common.ui import Tree

from ..sidebar_item import SideBarViewItem
from .menu import DirectoryContextMenu
from .placeholder import DirectoryTreePlaceholder
from .watcher import DirectoryTreeWatcher


class DirectoryTree(SideBarViewItem):
    """A view that displays the directory tree.

    The directory tree displays the contents of the active directory."""

    def __init__(
        self,
        master,
        startpath=None,
        observe_changes=False,
        itembar=True,
        style="",
        *args,
        **kwargs,
    ) -> None:
        self.title = "No folder opened"
        self.__actions__ = (
            (Icons.NEW_FILE, lambda: self.base.palette.show("newfile:")),
            (Icons.NEW_FOLDER, lambda: self.base.palette.show("newfolder:")),
            (Icons.REFRESH, self.refresh_root),
            (Icons.COLLAPSE_ALL, self.collapse_all),
        )
        super().__init__(master, itembar=itembar, *args, **kwargs)

        self.nodes = {}

        self.hide_dirs = [
            ".git",
        ]

        self.search_ignore_dirs = [
            ".git",
            "__pycache__",
            ".pytest_cache",
            "node_modules",
            "debug",
            "dist",
            "build",
        ]

        s = os.sep
        self.changes_ignore_dir_patterns = [
            f"*{s}.git{s}*",
            f"*{s}__pycache__{s}*",
            f"*{s}.pytest_cache{s}*",
            f"*{s}node_modules{s}*",
            f"*{s}debug{s}*",
            f"*{s}dist{s}*",
            f"*{s}build{s}*",
        ]

        self.ignore_exts = [".pyc"]
        self.git = self.base.git

        self.tree = Tree(
            self.content,
            startpath,
            doubleclick=self.openfile,
            singleclick=self.preview_file,
            style=style,
            *args,
            **kwargs,
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.tree.grid_remove()
        self.tree.bind("<<Open>>", self.toggle_node)

        self.placeholder = DirectoryTreePlaceholder(self.content)
        self.placeholder.grid(row=0, column=0, sticky=tk.NSEW)

        self.path = startpath
        self.watcher = DirectoryTreeWatcher(self, self.tree, observe_changes)
        self.loading = False

        self.ctxmenu = DirectoryContextMenu(self, "ExplorerContextMenu")
        self.tree.bind("<Button-3>", self.right_click)

        self.tree.tree.tag_configure("ignored", foreground=self.base.theme.disabled)

        if startpath:
            self.change_path(startpath)
        else:
            self.tree.insert("", 0, text="You have not yet opened a folder.")

    def right_click(self, e: tk.Event) -> None:
        """Shows the context menu on right click."""

        if item := self.tree.identify_row(e.y):
            self.tree.selection_set(item)
            self.tree.focus(item)
            self.ctxmenu.show(e)

    # IMPORTANT
    def change_path(self, path: str, create_root=True) -> None:
        """Changes the current directory and updates the treeview.
        Main interface for changing the current directory and updating the treeview."""

        self.nodes.clear()
        self.path = os.path.abspath(path) if path else path
        self.nodes[self.path] = ""
        if self.path:
            self.placeholder.grid_remove()
            self.tree.grid()
            self.tree.clear_tree()
            if create_root:
                self.create_root(self.path, subdir=False)
            self.watcher.watch()

            self.set_title(os.path.basename(self.path))
        else:
            self.tree.grid_remove()
            self.placeholder.grid()
            self.set_title("No folder opened")

    def create_root(self, path: str, parent="", subdir=True) -> None:
        """Creates the root node of the treeview."""

        if self.loading:
            return

        self.loading = True
        t = threading.Thread(
            target=self.run_create_sub_root if subdir else self.run_create_root,
            args=(path, parent),
        )
        t.daemon = True
        t.start()

    def run_create_root(self, path: str, parent="") -> None:
        """Updates the treeview with the contents of the given directory."""

        self.update_treeview(path, parent)
        self.loading = False

        self.base.statusbar.process_indicator.hide()

    def run_create_sub_root(self, path: str, parent="") -> None:
        """Updates the treeview with the contents of the given directory."""

        self.update_treeview(path, parent)
        self.loading = False

    def get_all_files(self) -> list:
        """Returns a list of all files in the treeview."""

        files = []
        for item in self.tree.get_children():
            if self.tree.item_type(item) == "file":
                files.append(
                    (
                        self.tree.item(item, "text"),
                        lambda _, item=item: print(self.tree.item_fullpath(item)),
                    )
                )

        return files

    def scandir(self, path: str) -> Iterator:
        """Scans the given directory and yields its contents."""

        for entry in os.scandir(path):
            isdir = os.path.isdir(entry.path)
            data = (
                entry.name,
                os.path.join(self.path, entry.path),
                isdir,
                entry.path.replace("\\", "/")
                + ("/" if isdir else ""),  # unix-like path
            )

            yield data

    def update_path(self, path: str) -> None:
        """Updates the treeview with the contents of the given directory."""

        if not path or any(path.endswith(i) for i in self.hide_dirs):
            return

        node = self.nodes.get(os.path.abspath(path))
        for i in self.tree.get_children(node):
            try:
                self.tree.delete(i)
            except:
                pass

        self.create_root(path, node)

    def update_treeview(self, parent_path: str, parent="") -> None:
        """Lazy loads the treeview with the contents of the given directory.

        Initially this was done recursively, but it was changed to a lazy load
        to improve performance. The treeview is updated only when the user
        expands a directory node."""

        if not os.path.exists(parent_path):
            return

        # sort: directories first, then files (alphabetic order)
        entries = sorted(self.scandir(parent_path), key=lambda x: (not x[2], x[0]))

        ignored = []
        if paths := [i[3] for i in entries]:
            ignored = self.git.ignore.check(paths)

        for name, path, isdir, unixlike in entries:
            if isdir:
                if name in self.hide_dirs:
                    continue

                try:
                    node = self.tree.insert(
                        parent or "",
                        "end",
                        text=f"  {name}",
                        values=[path, "directory"],
                        # image="foldericon",
                        open=False,
                        tags="ignored" if ignored and (unixlike in ignored) else "",
                    )
                    self.nodes[os.path.abspath(path)] = node
                    self.tree.insert(node, "end", text="loading...", tags="ignored")
                except:
                    self.refresh_root()
                    break

                # NOTE: recursive mode loading (not good for large projects)
                # self.update_treeview(path, node)
            else:
                if name.split(".")[-1] in self.ignore_exts:
                    continue

                try:
                    # TODO check filetype and get matching icon, cases
                    node = self.tree.insert(
                        parent,
                        "end",
                        text=f"  {name}",
                        values=[path, "file"],
                        # image="document",
                        tags="ignored" if ignored and (unixlike in ignored) else "",
                    )
                    self.nodes[os.path.abspath(path)] = node
                except:
                    self.refresh_root()
                    break

    def selected_directory(self) -> str:
        """Returns the selected directory path, or the current path if no directory is selected."""

        return (
            self.tree.selected_path().strip()
            if self.tree.selected_type() != "file"
            else self.tree.selected_parent_path().strip()
        ) or self.path

    def new_file(self, filename: str) -> None:
        """Creates a new file in the selected directory."""

        if not filename:
            return

        parent = (
            self.selected_directory()
            or self.base.active_directory
            or os.path.abspath(".")
        )
        path = os.path.join(parent, filename)

        if os.path.exists(path):
            # If user tries to create a new file with the name of an existing file
            # open that existing file in editor instead.
            self.base.open_editor(path)
            return

        with open(path, "w+") as f:
            f.write("")
        self.create_root(parent, self.nodes[parent])

    def new_folder(self, foldername: str) -> None:
        """Creates a new folder in the selected directory."""

        if not foldername:
            return

        parent = self.selected_directory()
        path = os.path.join(parent, foldername)
        try:
            os.makedirs(path, exist_ok=True)
        except:
            self.base.logger.error(
                f"Creating folder failed: no permission to write ('{path}')"
            )
            self.base.notifications.error("Creating folder failed: see logs")
            return
        self.create_root(parent, self.nodes[parent])

    def copy_path(self, *_) -> None:
        """Copies the absolute path of the selected item to the clipboard."""

        pyperclip.copy(self.tree.selected_path())

    def copy_relpath(self, *_) -> None:
        """Copies the relative path of the selected item to the clipboard."""

        pyperclip.copy(os.path.relpath(self.tree.selected_path(), self.path))

    def copy_name(self, *_) -> None:
        """Copies the name of the selected item to the clipboard."""

        pyperclip.copy(os.path.basename(self.tree.selected_path()))

    def copy_name_without_extension(self, *_) -> None:
        """Copies the name of the selected item without the extension to the clipboard."""

        pyperclip.copy(os.path.splitext(os.path.basename(self.tree.selected_path()))[0])

    def attach_to_chat(self, *_) -> None:
        """Attaches the selected item to the chat."""

        path = os.path.abspath(self.tree.selected_path())
        if self.tree.is_file_selected():
            self.base.ai.attach_file(path)

    def add_to_gitignore(self, *_) -> None:
        """Adds the selected item to the .gitignore file."""

        path = os.path.relpath(self.tree.selected_path(), self.path)
        self.git.ignore.add((path + "/") if self.tree.is_directory_selected() else path)

    def exclude_from_gitignore(self, *_) -> None:
        """Excludes the selected item from the .gitignore file."""

        path = os.path.relpath(self.tree.selected_path(), self.path)
        self.git.ignore.exclude(
            (path + "/") if self.tree.is_directory_selected() else path
        )

    def reopen_editor(self, *_) -> None:
        """Reopens the selected file in the editor."""

        path = os.path.abspath(self.tree.selected_path())
        if self.tree.is_file_selected():
            self.base.editorsmanager.reopen_editor(path)

    def reveal_in_explorer(self, *_) -> None:
        """Reveals the selected directory in the file explorer.

        This method is platform dependent, and uses the `subprocess` module
        to open the file explorer. If the file explorer executable is not found,
        it will log the error and show a warning notification."""

        path = self.selected_directory()
        try:
            if platform.system() == "Windows":
                subprocess.Popen(["start", path], shell=True)
            elif platform.system() == "Linux":
                try:
                    subprocess.Popen(["xdg-open", path])
                except OSError:
                    subprocess.Popen(["nautilus", path, "&"])
            else:
                subprocess.Popen(["open", path])
        except OSError as e:
            self.base.notifications.warning(
                "No File Explorer executable detected, see logs"
            )
            self.base.logger.error(
                f"Explorer: {e}\n(Mostly because no File explorer executable detected)"
            )
            return

    def open_in_terminal(self, *_) -> None:
        """Opens the selected directory in a terminal."""

        if path := self.selected_directory():
            self.base.terminalmanager.open_terminal(path)
            self.base.panel.show_panel()

    def rename_item(self, newname: str) -> None:
        """Renames the selected item in the treeview."""

        parent = self.tree.selected_parent_path()

        if path := self.tree.selected_path():
            shutil.move(
                path,
                os.path.join(self.tree.selected_parent_path() or self.path, newname),
            )

        self.update_path(parent)

    def delete_item(self) -> None:
        """Deletes the selected item from the treeview."""

        path = self.tree.selected_path()
        parent = self.tree.selected_parent_path()
        if not askyesno("Delete", f"Are you sure of deleting {path}?"):
            return

        try:
            if self.tree.selected_type() == "directory":
                shutil.rmtree(path)
                return
            os.remove(path)
        except OSError as e:
            self.base.notifications.warning("Removing failed, see logs")
            self.base.logger.error(f"Removing failed: {e}")
            return

        # refresh parent of the deleted item
        self.update_path(parent)

    def collapse_all(self, *_) -> None:
        """Collapses all nodes in the treeview."""

        for node in self.tree.get_children(""):
            self.tree.item(node, open=False)

    def refresh_selected_parent(self, *_) -> None:
        """Reloads entire parent node of the selected node."""

        self.update_path(self.tree.selected_parent_path())

    def refresh_root(self, *_) -> None:
        """Reloads entire treeview from the root."""

        self.update_path(self.path)

    def close_directory(self) -> None:
        """Closes the current directory and clears the treeview."""

        self.change_path(None)

    def toggle_node(self, *_) -> None:
        """Toggles the selected node, if it's a directory."""

        node = self.tree.focus()
        for i in self.tree.get_children(node):
            self.tree.delete(i)

        self.create_root(self.tree.selected_path(), node)

    def openfile(self, _) -> None:
        """Opens the selected file in an editor."""

        if self.tree.selected_type() != "file":
            return

        path = self.tree.selected_path()
        self.base.open_editor(path)

    def preview_file(self, _) -> None:
        # TODO preview editors -- extra preview param for editors
        return



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame

from .kind import Kind

if typing.TYPE_CHECKING:
    from biscuit.language.data import Completion

    from . import AutoComplete


class CompletionItem(Frame):
    def __init__(self, master: AutoComplete, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.config(width=500, **self.base.theme.editors.autocomplete)
        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.editors.autocomplete.item.values()
        )

        self.selected = False
        self.hovered = False

        self.replace_start: str = ""
        self.replace_end: str = ""
        self.replace_text: str = ""
        self.documentation: str = ""
        # TODO add GUI for documentation

        self.kind = Kind(self)
        self.text = tk.Text(
            self,
            font=self.base.settings.font,
            fg=self.fg,
            bg=self.bg,
            cursor="hand2",
            relief=tk.FLAT,
            highlightthickness=0,
            width=30,
            height=1,
        )

        self.text.tag_config(
            "term",
            foreground=self.base.theme.biscuit,
            font=self.base.settings.font_bold,
        )
        self.text.config(state=tk.DISABLED)

        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.kind.grid(row=0, column=0, sticky=tk.NSEW)
        self.text.grid(row=0, column=1, sticky=tk.NSEW)

        self.kind.bind("<Button-1>", self.on_click)
        self.text.bind("<Button-1>", self.on_click)
        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

    def set_data(self, word: str, term: str):
        self.replace_start = ""
        self.replace_end = ""
        self.replace_text = word
        self.documentation = ""

        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.insert(tk.END, word)
        self.text.config(state=tk.DISABLED)

        self.kind.set_kind()
        self.mark_term(word, term)

    def lsp_set_data(self, completion: Completion, term: str):
        self.replace_start = completion.replace_start
        self.replace_end = completion.replace_end
        self.replace_text = completion.replace_text
        self.documentation = completion.documentation

        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.insert(tk.END, completion.display_text)
        self.text.config(state=tk.DISABLED)

        self.kind.set_kind(completion.kind)
        self.mark_term(completion.display_text, term)

    def clear_data(self):
        self.replace_start = ""
        self.replace_end = ""
        self.replace_text = ""
        self.documentation = ""

        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.config(state=tk.DISABLED)

        self.kind.set_kind()
        self.clear_mark()

    def clear_mark(self):
        self.text.tag_remove("term", 1.0, tk.END)

    def mark_term(self, display_text: str, term: str):
        self.clear_mark()

        # Prioritize case sensitive match
        start_pos = display_text.find(term)
        if start_pos == -1:
            # Case insensitive match
            start_pos = display_text.lower().find(term)
            if start_pos == -1:
                return

        end_pos = start_pos + len(term)
        self.text.tag_add("term", f"1.{start_pos}", f"1.{end_pos}")

    def on_click(self, *_):
        self.master.choose(this=self)

    def on_hover(self, *_):
        if not self.selected:
            self.kind.config(bg=self.hbg)
            self.text.config(bg=self.hbg)
            self.hovered = True

    def off_hover(self, *_):
        if not self.selected:
            self.kind.config(bg=self.bg)
            self.text.config(bg=self.bg)
            self.hovered = False

    def toggle_selection(self):
        if self.selected:
            self.select()
        else:
            self.deselect()

    def select(self):
        self.kind.config(bg=self.hbg)
        self.text.config(bg=self.hbg, fg=self.hfg)
        self.selected = True

    def deselect(self):
        self.kind.config(bg=self.bg)
        self.text.config(bg=self.bg, fg=self.fg)
        self.selected = False



"""Shell classes for the terminal view."""

import platform

from .bash import Bash
from .cmd import CommandPrompt
from .default import Default
from .powershell import PowerShell

SHELLS = [Default, PowerShell]

if platform.system() == "Windows":
    SHELLS.append(CommandPrompt)
elif platform.system() == "Linux":
    SHELLS.append(Bash)

# def get_shells(base):
#     return [(i, lambda i=i: base.open_shell(i)) for i in shells.keys()]

# def get_shell(name):
#     return shells.get(name, Default)

# def register_shell(shell):
#     global shells
#     try:
#         shells[shell.name] = shell
#     except AttributeError:
#         shells[f"shell {len(shells) + 1}"] = shell



"""Layout

Contains the layout of the application. It is divided into 3 main parts:

Root
├── Menubar
├── <Container>
│  ├── Sidebar (Explorer, Search, Source Control, etc.)
│  └── Content
│     ├── Editors
│     └── Panel (Terminals, Logs, Problems, etc.)
└── StatusBar

The layout is designed to be modular and flexible. Each part can be easily replaced or modified.
Extensions can easily access and modify the layout.
"""

from .root import *



from __future__ import annotations

import tkinter as tk
import typing
from hashlib import md5

from biscuit.common.icons import Icons
from biscuit.common.ui import Scrollbar

from ..editorbase import BaseEditor
from .linenumbers import LineNumbers
from .menu import RunMenu
from .minimap import Minimap
from .text import Text

if typing.TYPE_CHECKING:
    from tkinter.font import Font

    from biscuit.debugger.base import DebuggerBase
    from biscuit.editor.editor import Editor


class TextEditor(BaseEditor):
    def __init__(
        self,
        master: Editor,
        path="",
        exists=True,
        language="",
        minimalist=False,
        standalone=False,
        load_file=True,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, path, exists, *args, **kwargs)
        self.font: Font = self.base.settings.font
        self.path = path
        self.exists = exists
        self.standalone = standalone
        self.minimalist = minimalist or self.standalone
        self.language = language
        self.editable = True
        self.run_command_value = ""
        self.debugger: DebuggerBase = None
        self.runmenu = None
        self.unsupported = False
        self.content_hash = ""

        if not self.standalone:
            self.__buttons__ = [
                (Icons.REFRESH, self.base.editorsmanager.reopen_active_editor),
            ]

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

        self.linenumbers = LineNumbers(self)
        self.scrollbar = Scrollbar(self, orient=tk.VERTICAL, style="EditorScrollbar")

        if not self.minimalist:
            self.minimap = Minimap(self)
            self.minimap.grid(row=0, column=2, sticky=tk.NS)

        self.text = Text(
            self,
            path=self.path,
            exists=self.exists,
            minimalist=self.minimalist,
            standalone=self.standalone,
            language=self.language,
        )
        self.language = self.text.language

        if self.exists:
            if load_file:
                self.text.load_file()

            self.text.update_idletasks()

            if not self.standalone:
                self.run_command_value = self.base.execution_manager.get_command(self)
                self.__buttons__.insert(
                    0,
                    {
                        "icon": Icons.PLAY,
                        "event": lambda: self.run_file(),
                        "width": 1,
                        "hfg_only": True,
                    },
                )

                self.runmenu = RunMenu(self, "run menu")
                if self.run_command_value:
                    self.runmenu.add_command(
                        f"Run {self.language} file", lambda: self.run_file()
                    )
                    self.runmenu.add_separator()
                self.runmenu.add_command(
                    "Run in dedicated terminal", lambda: self.run_file(dedicated=True)
                )
                self.runmenu.add_command(
                    "Run in external console", lambda: self.run_file(external=True)
                )
                self.runmenu.add_separator()
                self.runmenu.add_command(
                    "Configure Run...",
                    lambda: self.base.commands.show_run_config_palette(
                        self.run_command_value
                    ),
                )

                self.__buttons__.insert(
                    1,
                    {
                        "icon": Icons.CHEVRON_DOWN,
                        "event": self.runmenu.show,
                        "iconsize": 6,
                        "width": 1,
                    },
                )

                self.debugger = self.base.debugger_manager.request_debugger_for_editor(
                    self
                )
                if self.debugger:
                    self.__buttons__.insert(2, (Icons.DEBUG_ALT, self.run_debugger))
                    self.runmenu.add_separator()
                    self.runmenu.add_command(
                        f"Debug {self.language} file", self.run_debugger
                    )

                    if self.path in self.debugger.breakpoints:
                        self.linenumbers.breakpoints = self.debugger.breakpoints[
                            self.path
                        ]
                        self.linenumbers.redraw()

        self.linenumbers.attach(self.text)
        if not self.minimalist:
            self.minimap.attach(self.text)
        self.scrollbar.config(command=self.text.yview)

        self.text.config(font=self.font)
        self.text.configure(yscrollcommand=self.scrollbar.set)

        self.linenumbers.grid(row=0, column=0, sticky=tk.NS)
        self.text.grid(row=0, column=1, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, column=3, sticky=tk.NS)

        self.text.bind("<<Change>>", self.on_change)
        self.text.bind("<<Scroll>>", self.on_scroll)

        self.on_change()
        self.on_scroll()

        if self.base.settings.config.auto_save_enabled:
            self.auto_save()

    def run_debugger(self) -> None:
        if self.debugger:
            self.debugger.launch_standalone(self)

    def update_breakpoints(self, breakpoints: set[int]) -> None:
        if self.debugger:
            self.debugger.update_breakpoints(self.path, breakpoints)

    def __getattr__(self, name) -> typing.Any:
        """For methods that are not in this class, assume it is present in Text widget
        and delegate the call to the `Text` widget."""

        text_methods = set(dir(self.text))
        if name in text_methods:
            attr = getattr(self.text, name)
            if self.editable:
                return attr
            else:
                return lambda *args, **kwargs: None
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def file_loaded(self) -> None:
        self.recalculate_content_hash()
        self.event_generate("<<FileLoaded>>", when="tail")
        self.text.event_generate("<<FileLoaded>>", when="tail")

    def recalculate_content_hash(self) -> None:
        """Recalculate the hash of the editor content"""
        self.content_hash = self.calculate_content_hash()

    def calculate_content_hash(self) -> str | None:
        """Calculate the hash of the editor content"""
        if self.exists and self.editable:
            text = self.text.get_all_text()
            return md5(text.encode()).hexdigest()

    @property
    def breakpoints(self):
        return self.linenumbers.breakpoints

    @property
    def unsaved_changes(self):
        """Check if the editor content has changed"""
        if self.editable:
            if not self.content_hash:
                return False

            return self.content_hash != self.calculate_content_hash()

    def run_file(self, dedicated=False, external=False) -> None:
        if not self.run_command_value:
            self.base.notifications.warning(
                "No programs are configured to run this file."
            )
            self.base.commands.show_run_config_palette(self.run_command_value)
            return

        self.save()

        # add another dedicated terminal if there is an active terminal
        if self.base.terminalmanager.active_terminal and dedicated:
            self.base.terminalmanager.add_default_terminal()

        if not external:
            self.base.panel.show_terminal()
        self.base.execution_manager.run_command(self, external=external)

    def set_run_command(self, command) -> None:
        self.run_command_value = command
        self.run_file()

    def on_change(self, *_) -> None:
        self.linenumbers.redraw()
        try:
            if not self.standalone:
                self.base.update_statusbar()
        except ValueError:
            pass
        self.text.refresh()
        if not self.minimalist:
            self.minimap.redraw_cursor()
        self.event_generate("<<Change>>")

    def on_scroll(self, *_) -> None:
        self.linenumbers.redraw()
        if not self.minimalist:
            self.minimap.redraw()
        self.event_generate("<<Scroll>>")

    def unsupported_file(self) -> None:
        self.unsupported = True
        self.text.show_unsupported_dialog()
        self.linenumbers.grid_remove()
        self.scrollbar.grid_remove()
        self.editable = False

    def focus(self) -> None:
        self.text.focus()
        self.on_change()

    def set_fontsize(self, size) -> None:
        self.font.configure(size=size)
        self.linenumbers.set_bar_width(size * 3)
        self.on_change()

    def save(self, path=None) -> None:
        if self.editable:
            self.recalculate_content_hash()
            self.text.save_file(path)

    def auto_save(self) -> None:
        if self.standalone:
            return

        self.save()
        self.base.after(self.base.settings.config.auto_save_timer_ms, self.auto_save)



from __future__ import annotations

import typing

from .assistant import Assistant
from .commands import Commands
from .editors import Editors
from .logger import Logger
from .notifications import Notifications
from .releases import Releases
from .views import Views

if typing.TYPE_CHECKING:
    from biscuit import App


class ExtensionsAPI:
    """Extensions API

    This class is used to interact with the Biscuit API.
    The API provides access to the `base` (App) instance which means
    the extension can interact with every part of the application without
    any restrictions.

    Furthermore, the API provides access to the following endpoints for easier access:
    - `commands`
    - `logger`
    - `editors`
    - `notifications`
    - `utils`
    - `views`
    - `releases`

    REGISTER METHODS:
    - `register_comment_prefix` - Register a comment prefix for a language
    - `register_game` - Register a game
    - `register_langserver` - Register a language server
    - `register_run_command` - Register a run command
    """

    def __init__(self, base: App) -> None:
        self.base = base

        self.menubar = self.base.menubar
        self.statusbar = self.base.statusbar
        self.sidebar = self.base.sidebar
        self.panel = self.base.panel
        self.sysinfo = self.base.system
        self.editorsmanager = self.base.editorsmanager
        self.terminalmanager = self.base.terminalmanager
        self.languageservermanager = self.base.language_server_manager

        self.commands = Commands(self.base)
        self.logger = Logger(self.base)
        self.editors = Editors(self.base)
        self.notifications = Notifications(self.base)
        self.views = Views(self.base)
        self.releases = Releases(self.base)
        self.assistant = Assistant(self.base)

        self.register_comment_prefix = self.base.register_comment_prefix
        self.register_game = self.base.register_game
        self.register_langserver = self.base.register_langserver
        self.register_run_command = self.base.register_run_command

    def register(self, name: str, extension: object) -> None:
        """Register an extension"""

        self.base.extensions_manager.register_installed(name, extension)

    def register_extension(self, name: str, extension: object) -> None:
        """Register an extension"""

        self.register(name, extension)



import tkinter

# from tkinter.dnd import *


class DndHandler:

    root = None

    def __init__(self, source, event):
        if event.num > 5:
            return
        root = event.widget._root()
        try:
            root.__dnd
            return # Don't start recursive dnd
        except AttributeError:
            root.__dnd = self
            self.root = root
        self.source = source
        self.target = None
        self.initial_button = button = event.num
        self.initial_widget = widget = event.widget
        self.release_pattern = "<B%d-ButtonRelease-%d>" % (button, button)
        self.save_cursor = widget['cursor'] or ""
        widget.bind(self.release_pattern, self.on_release)
        widget.bind("<Motion>", self.on_motion)
        widget['cursor'] = "hand2"

    def __del__(self):
        root = self.root
        self.root = None
        if root is not None:
            try:
                del root.__dnd
            except AttributeError:
                pass

    def on_motion(self, event):
        x, y = event.x_root, event.y_root
        target_widget = self.initial_widget.winfo_containing(x, y)
        source = self.source
        new_target = None
        while target_widget is not None:
            try:
                attr = target_widget.dnd_accept
            except AttributeError:
                pass
            else:
                new_target = attr(source, event)
                if new_target is not None:
                    break
            target_widget = target_widget.master
        old_target = self.target
        if old_target is new_target:
            if old_target is not None:
                old_target.dnd_motion(source, event)
        else:
            if old_target is not None:
                self.target = None
                old_target.dnd_leave(source, event)
            if new_target is not None:
                new_target.dnd_enter(source, event)
                self.target = new_target

    def on_release(self, event):
        self.finish(event, 1)

    def cancel(self, event=None):
        self.finish(event, 0)

    def finish(self, event, commit=0):
        target = self.target
        source = self.source
        widget = self.initial_widget
        root = self.root
        try:
            del root.__dnd
            self.initial_widget.unbind(self.release_pattern)
            self.initial_widget.unbind("<Motion>")
            widget['cursor'] = self.save_cursor
            self.target = self.source = self.initial_widget = self.root = None
            if target is not None:
                if commit:
                    target.dnd_commit(source, event)
                else:
                    target.dnd_leave(source, event)
        finally:
            source.dnd_end(target, event)


def dnd_start(source, event):
    h = DndHandler(source, event)
    if h.root is not None:
        return h
    else:
        return None
    
# Now below is the sample application using tkinter.dnd

class Tester:
    def __init__(self, root):
        self.top = tkinter.Toplevel(root)
        self.canvas = tkinter.Canvas(self.top, width=100, height=100)
        self.canvas.pack(fill="both", expand=1)
        self.canvas.dnd_accept = self.dnd_accept
        self.top.minsize(300, 150)

    def dnd_accept(self, source, event):
        return self

    def dnd_enter(self, source, event):
        self.canvas.focus_set()
        x, y = source.where(self.canvas, event)
        x1, y1, x2, y2 = source.canvas.bbox(source.id)
        dx, dy = x2-x1, y2-y1
        self.dndid = self.canvas.create_rectangle(x, y, x+dx, y+dy)
        self.dnd_motion(source, event)

    def dnd_motion(self, source, event):
        x, y = source.where(self.canvas, event)
        x1, y1, x2, y2 = self.canvas.bbox(self.dndid)
        self.canvas.move(self.dndid, x-x1, y-y1)

    def dnd_leave(self, source, event):
        self.top.focus_set()
        self.canvas.delete(self.dndid)
        self.dndid = None

    def dnd_commit(self, source, event):
        self.dnd_leave(source, event)
        x, y = source.where(self.canvas, event)
        source.attach(self.canvas, x, y)

class Icon:

    def __init__(self, name):
        self.name = name
        self.canvas = self.label = self.id = None

    def attach(self, canvas, x=10, y=10):
        if canvas is self.canvas:
            self.canvas.coords(self.id, x, y)
            return
        if self.canvas is not None:
            self.detach()
        if canvas is None:
            return
        label = tkinter.Label(canvas, text=self.name,
                              borderwidth=2, relief="raised")
        id = canvas.create_window(x, y, window=label, anchor="nw")
        self.canvas = canvas
        self.label = label
        self.id = id
        label.bind("<ButtonPress>", self.press)

    def detach(self):
        canvas = self.canvas
        if canvas is None:
            return
        id = self.id
        label = self.label
        self.canvas = self.label = self.id = None
        canvas.delete(id)
        label.destroy()

    def press(self, event):
        if dnd_start(self, event):
            # where the pointer is relative to the label widget:
            self.x_off = event.x
            self.y_off = event.y
            # where the widget is relative to the canvas:
            self.x_orig, self.y_orig = self.canvas.coords(self.id)

    def move(self, event):
        x, y = self.where(self.canvas, event)
        self.canvas.coords(self.id, x, y)

    def putback(self):
        self.canvas.coords(self.id, self.x_orig, self.y_orig)

    def where(self, canvas, event):
        # where the corner of the canvas is relative to the screen:
        x_org = canvas.winfo_rootx()
        y_org = canvas.winfo_rooty()
        # where the pointer is relative to the canvas widget:
        x = event.x_root - x_org
        y = event.y_root - y_org
        # compensate for initial pointer offset
        return x - self.x_off, y - self.y_off

    def dnd_end(self, target, event):

        # if the item is dropped to nowhere, create a new window with that item info
        if not target:
            t = Tester(self.canvas)
            t.top.title(self.name)
            t.top.geometry(f"+{event.x_root}+{event.y_root}")
            t.dnd_enter(self, event)
            t.dnd_commit(self, event)

root = tkinter.Tk()
root.geometry("+1+1")
tkinter.Button(command=root.quit, text="Quit").pack()
t1 = Tester(root)
t1.top.geometry("+1+60")
t2 = Tester(root)
t2.top.geometry("+220+60")
t3 = Tester(root)
t3.top.geometry("+340+60")
i1 = Icon("ICON1")
i2 = Icon("ICON2")
i3 = Icon("ICON3")
i1.attach(t1.canvas)
i2.attach(t2.canvas)
i3.attach(t3.canvas)
root.mainloop()



import tkinter as tk
from typing import Any

from biscuit.common.icons import Icons

from ..editorbase import BaseEditor
from ..text import TextEditor
from .renderer import HTMLRenderer


class HTMLEditor(BaseEditor):
    def __init__(self, master, path, exists=False, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.path = path
        self.exists = exists
        self.editable = True
        self.preview_enabled = False

        # TODO open in browser button

        self.__buttons__ = ((Icons.OPEN_PREVIEW, self.toggle_preview),)

        self.editor = TextEditor(self, path, exists=exists)
        self.editor.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 1))

        self.renderer = HTMLRenderer(self, editor=self.editor)

        self.left = self.text = self.editor.text
        self.right = self.renderer.text

        self.editor.scrollbar["command"] = self.on_scrollbar
        self.renderer.scrollbar["command"] = self.on_scrollbar
        self.left["yscrollcommand"] = self.on_textscroll
        # self.right['yscrollcommand'] = self.on_textscroll

        self.editor.bind("<<Change>>", self.on_change)
        self.edit_undo = self.editor.edit_undo
        self.editor_redo = self.editor.edit_redo

    @property
    def unsaved_changes(self) -> bool:
        return self.editor.unsaved_changes

    def toggle_preview(self, *_):
        if self.preview_enabled:
            self.grid_columnconfigure(1, weight=0)
            self.renderer.grid_forget()
        else:
            self.grid_columnconfigure(1, weight=1)
            self.renderer.grid(row=0, column=1, sticky=tk.NSEW)
        self.preview_enabled = not self.preview_enabled
        self.on_change()

    def on_scrollbar(self, *args) -> None:
        self.left.yview(*args)
        self.editor.on_scroll()
        self.right.yview(*args)

    def on_textscroll(self, *args) -> None:
        self.editor.scrollbar.set(*args)
        self.renderer.scrollbar.set(*args)
        self.on_scrollbar("moveto", args[0])

    def on_change(self, *_):
        if self.preview_enabled:
            self.renderer.refresh()



import random
import tkinter as tk
from tkinter import messagebox

from biscuit.common.ui import Button, Frame, IconButton, Label

from .gamebase import BaseGame

# Game constants
BOARD_SIZE = 10
NUM_MINES = 10


class Tile(IconButton):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.config(
            font=("Segoi UI", 13),
            width=2,
            bg=self.base.theme.biscuit,
            fg="white",
            activebackground=self.base.theme.biscuit_dark,
        )

    def reveal_count(self, count) -> None:
        self.config(font=("Segoi UI", 13), text=count)
        bg = self.base.theme.editors.background
        self.config(bg=bg, activebackground=bg)

    def reveal_icon(self, icon, **kw) -> None:
        self.config(font=("codicon", 13), **kw)
        self.set_icon(icon)

    def clear(self) -> None:
        self.config(fg=self.base.theme.editors.background)


class Minesweeper(BaseGame):
    name = "Minesweeper!"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.container = Frame(self, bg=self.base.theme.border, padx=1, pady=1)
        self.container.pack(pady=100)

        self.gameover_frame = Frame(self, **self.base.theme.editors, pady=50)
        Label(
            self.gameover_frame,
            text="Game Over",
            font=("Fixedsys", 20),
            fg=self.base.theme.biscuit,
            **self.base.theme.editors
        ).pack()
        restart = Button(self.gameover_frame, "Retry!", self.reload)
        restart.config(font=("Fixedsys", 20))
        restart.pack(pady=20)

        self.start_game()

    def reload(self, *_):
        self.gameover_frame.pack_forget()
        self.container.pack(pady=100)
        self.start_game()

    def start_game(self):
        self.mine_positions = []
        self.board = [[0] * BOARD_SIZE for _ in range(BOARD_SIZE)]
        self.buttons = [[None] * BOARD_SIZE for _ in range(BOARD_SIZE)]

        self.generate_mines()
        self.create_tiles()

    def generate_mines(self):
        self.mine_positions = random.sample(range(BOARD_SIZE * BOARD_SIZE), NUM_MINES)

        for position in self.mine_positions:
            row = position // BOARD_SIZE
            col = position % BOARD_SIZE
            self.board[row][col] = -1

    def create_tiles(self):
        for row in range(BOARD_SIZE):
            for col in range(BOARD_SIZE):
                button = Tile(self.container, "", width=2, height=1)
                button.grid(
                    row=row, column=col, padx=(0, 1), pady=(0, 1), sticky=tk.NSEW
                )
                button.bind(
                    "<Button-1>", lambda e, r=row, c=col: self.button_click(e, r, c)
                )
                button.bind(
                    "<Button-3>", lambda e, r=row, c=col: self.button_click(e, r, c)
                )
                self.buttons[row][col] = button

    def button_click(self, event, row, col):
        # Handle left-click
        if event.num == 1:
            if self.board[row][col] == -1:
                self.buttons[row][col].reveal_icon("twitter", bg="red")
                self.game_over()
            else:
                count = self.count_adjacent_mines(row, col)
                self.buttons[row][col].reveal_count(count)
                self.buttons[row][col].unbind("<Button-1>")
                if count == 0:
                    self.buttons[row][col].clear()
                    self.reveal_empty_cells(row, col)
        # Handle right-click
        elif event.num == 3:
            # Check if the button is already flagged
            if self.buttons[row][col]["text"] == "pinned":
                # Remove the flag
                self.buttons[row][col].config(text="")
                self.buttons[row][col].bind(
                    "<Button-1>", lambda e, r=row, c=col: self.button_click(e, r, c)
                )
            else:
                # Flag the button
                self.buttons[row][col].reveal_icon("pinned")
                self.buttons[row][col].unbind("<Button-1>")

    def count_adjacent_mines(self, row, col):
        count = 0
        for i in range(max(0, row - 1), min(row + 2, BOARD_SIZE)):
            for j in range(max(0, col - 1), min(col + 2, BOARD_SIZE)):
                if self.board[i][j] == -1:
                    count += 1
        return count

    def reveal_empty_cells(self, row, col):
        for i in range(max(0, row - 1), min(row + 2, BOARD_SIZE)):
            for j in range(max(0, col - 1), min(col + 2, BOARD_SIZE)):
                if self.buttons[i][j]["text"] == "" and self.board[i][j] != -1:
                    count = self.count_adjacent_mines(i, j)
                    self.buttons[i][j].reveal_count(count)
                    self.buttons[i][j].unbind("<Button-1>")
                    if count == 0:
                        self.buttons[i][j].clear()
                        self.reveal_empty_cells(i, j)

    def game_over(self):
        for position in self.mine_positions:
            row = position // BOARD_SIZE
            col = position % BOARD_SIZE
            if self.buttons[row][col]["text"] != "twitter":
                self.buttons[row][col].reveal_icon("twitter")
                self.buttons[row][col].unbind("<Button-1>")

        self.container.pack_forget()
        self.gameover_frame.pack(pady=100)



import os
import tkinter as tk

from biscuit.common import ActionSet
from biscuit.common.icons import Icons

from ..sidebar_view import SideBarView
from .directorytree import DirectoryTree
from .menu import ExplorerMenu


class Explorer(SideBarView):
    """A view that displays the file explorer.

    - The explorer view displays the directory tree and open editors.
    - context menu that allows the user to manage the files in the directory.
    - The user can search, manage files in the directory using palette
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = []
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.HOME
        self.name = "Explorer"

        # self.menu = ExplorerMenu(self, "files")
        # self.menu.add_command("Search", self.base.commands.search_files)
        # self.add_action(Icons.ELLIPSIS, self.menu.show)

        self.top.pack_forget()

        self.directory = DirectoryTree(self, observe_changes=True)
        self.add_item(self.directory)

        self.filesearch_actionset = ActionSet("Search files", "file:", [])

        self.newfile_actionset = ActionSet(
            "Add new file to directory",
            "newfile:",
            pinned=[
                [
                    "Create new file: {}",
                    lambda filename=None: self.directory.new_file(filename),
                ]
            ],
        )
        self.base.palette.register_actionset(lambda: self.newfile_actionset)

        self.newfolder_actionset = ActionSet(
            "Add new folder to parent directory",
            "newfolder:",
            pinned=[
                [
                    "Create new folder: {}",
                    lambda foldername=None: self.directory.new_folder(foldername),
                ]
            ],
        )
        self.base.palette.register_actionset(lambda: self.newfolder_actionset)

        self.rename_actionset = ActionSet(
            "Rename a file/folder",
            "rename:",
            pinned=[
                [
                    "Rename to: {}",
                    lambda newname=None: self.directory.rename_item(newname),
                ]
            ],
        )
        self.base.palette.register_actionset(lambda: self.rename_actionset)

    def get_actionset(self, term: str) -> ActionSet:
        self.filesearch_actionset.update(self.filesearch(term))
        return self.filesearch_actionset

    def filesearch(self, t):
        if not self.base.active_directory:
            return []

        pys = []
        for r, d, f in os.walk(self.base.active_directory):
            if any(i in r for i in self.directory.search_ignore_dirs):
                d[:] = []
                continue
            for file in f:
                if t in file:
                    pys.append(
                        (
                            file,
                            lambda _, path=os.path.join(r, file): self.base.open_editor(
                                path
                            ),
                        )
                    )
        return pys



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui.bubble import Bubble
from biscuit.common.ui.icon import IconButton


class NotificationBubble(Bubble):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

    def get_pos(self, *_):
        return (
            f"+{self.master.winfo_rootx() + (self.master.winfo_width() - self.winfo_width()) // 2}"
            + f"+{self.master.winfo_rooty() + self.master.winfo_height() - 2}"
        )


class Notifications(IconButton):
    """Notification

    - Notification indicator
    - Shows the number of notifications
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(
            master, Icons.BELL, hfg_only=True, padx=10, pady=8, *args, **kwargs
        )

        self.base.notifications.set_button(self)
        self.event = self.base.notifications.show

        self.bubble = NotificationBubble(self, "No notifications")
        self.bind("<Enter>", self.bubble.show)
        self.bind("<Leave>", self.bubble.hide)



from __future__ import annotations

import typing
from pathlib import Path

import toml

if typing.TYPE_CHECKING:
    from .manager import DebuggerManager

CONFIG_FILE = "launch.toml"


class ConfigLoader:
    def __init__(self, manager: DebuggerManager):
        self.manager = manager
        self.base = manager.base
        self.config_found = False
        self.configs: list[dict] = []
        self.path: Path = None

        self.base.bind("<<DirectoryChanged>>", self.check, add=True)

    def check(self, *_) -> None:
        self.config_found = False
        if not self.base.active_directory:
            self.refresh_debugger()
            return

        self.path = Path(self.base.active_directory) / ".biscuit" / CONFIG_FILE
        if not self.path.exists():
            self.refresh_debugger()
            return

        self.config_found = True
        self.refresh_debugger()

        self.load(self.path)

    def refresh_debugger(self):
        self.base.debug.refresh()

    def load(self, path: Path) -> None:
        with open(path, "r") as f:
            data = toml.load(f)

        self.configs = data.values()
        self.base.debug.refresh_configs()



from __future__ import annotations

import typing

from biscuit.common.actionset import ActionSet
from biscuit.language import Languages

from .base import DebuggerBase
from .config import ConfigLoader
from .python import PythonDebugger

if typing.TYPE_CHECKING:
    from biscuit import App
    from biscuit.editor.text import TextEditor


class DebuggerManager:
    """Debugger manager class.
    This class manages the debugger instances and provides methods to run them."""

    def __init__(self, base: App):
        self.base = base

        self.debuggers: dict[str, type[DebuggerBase]] = (
            {}
        )  # language -> debugger type mapping
        self.debuggers[Languages.PYTHON] = PythonDebugger

        self.spawned: dict[str, DebuggerBase] = (
            {}
        )  # language -> active debugger instance

        self._latest: DebuggerBase = None

        self.config_loader = ConfigLoader(self)

    def register_actionsets(self) -> None:
        """Register the debugger action sets."""

        set_local_actionset = ActionSet(
            "debugger: set local variable",
            "debugger.set_local",
            pinned=[
                ["Change value to {}", self.base.debug.variables.set_variable_callback]
            ],
        )
        self.base.register_actionset(lambda: set_local_actionset)

    @property
    def latest(self) -> DebuggerBase:
        """Get the latest debugger instance
        Returns:
            DebuggerBase: the latest debugger instance
        """
        return self._latest

    @latest.setter
    def latest(self, debugger: DebuggerBase) -> None:
        """Set the latest debugger instance
        Args:
            debugger (DebuggerBase): the debugger instance
        """
        self._latest = debugger

    def request_debugger_for_editor(self, editor: TextEditor) -> DebuggerBase:
        """Get or create a debugger instance for the given editor.
        Args:
            editor (TextEditor): the editor instance
        Returns:
            DebuggerBase: the debugger instance
        """
        return self.request_debugger(editor.language)

    def request_debugger(self, language: str) -> DebuggerBase:
        """Get or create a debugger instance for the given editor's language.
        Args:
            editor (TextEditor): the editor instance
        Returns:
            DebuggerBase: the debugger instance
        """
        if not (language and self.is_debugger_available(language)):
            return

        language = language.lower()
        if language not in self.spawned:
            debugger_type = self.debuggers.get(language)
            if not debugger_type:
                return

            self.spawned[language] = debugger_type(self)
        return self.spawned[language]

    def is_debugger_available_for_editor(self, editor: TextEditor) -> bool:
        """Check if a debugger is available for the given editor.
        Args:
            editor (TextEditor): the editor instance
        Returns:
            bool: True if a debugger is available, False otherwise
        """
        return self.is_debugger_available(editor.language)

    def is_debugger_available(self, language: str) -> bool:
        """Check if a debugger is available for the given editor's language.
        Args:
            editor (TextEditor): the editor instance
        Returns:
            bool: True if a debugger is available, False otherwise
        """
        return language.lower() in self.debuggers

    def register_debugger(self, language: str, debugger: type[DebuggerBase]) -> None:
        """Register a debugger for a specific language.
        Args:
            language (str): the language
            debugger (type[DebuggerBase]): the debugger type
        """
        self.debuggers[language] = debugger

    def unregister_debugger(self, language: str) -> None:
        """Unregister a debugger for a specific language.
        Args:
            language (str): the language
        """
        self.debuggers.pop(language, None)



from biscuit.common import Menu


class EditorsbarMenu(Menu):
    def get_coords(self, e) -> tuple[int, int]:
        return (
            e.widget.winfo_rootx() + e.widget.winfo_width() - self.winfo_width(),
            e.widget.winfo_rooty() + e.widget.winfo_height(),
        )



import inspect
from functools import wraps

__all__ = ["command_palette_ignore", "extract_commands", "formalize_command"]


def command_palette_ignore(func):
    """Decorator that marks a method will be marked during the method
    extraction process for command palette."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper._ignored = True
    return wrapper


def predicate(obj):
    """Ignore built-in methods, other data"""

    return inspect.ismethod(obj) and not inspect.isbuiltin(obj)


def extract_commands(cls):
    """Extracts all methods from a class for command palette."""

    filtered = []
    methods = inspect.getmembers(cls, predicate=predicate)
    for name, method in methods:
        if not name.startswith("__") and not getattr(method, "_ignored", False):
            # TODO: make use of the documentation of commands in command palette
            # doc = inspect.getdoc(method) or "No documentation available."
            filtered.append((name, method))

    return filtered


def formalize_command(command: str) -> str:
    """Formalizes a command string to be used in the command palette."""

    return command.replace("_", " ").capitalize()



from __future__ import annotations

import typing

import toml

from .workspace import Workspace

if typing.TYPE_CHECKING:
    from .manager import WorkspaceManager


class WorkspaceLoader:
    def __init__(self, manager: WorkspaceManager):
        self.manager = manager
        self.base = manager.base

    def load(self, path):
        with open(path) as f:
            data = toml.load(f)
            self.open_workspace(Workspace(self, path, data["dirs"]))

    def dump_modified(self, path: str, dirs: list[str]):
        with open(path, "w") as f:
            toml.dump({"dirs": dirs}, f)

    def open_workspace(self, workspace: Workspace):
        self.manager.set_workspace(workspace)

    def save(self, active_workspace: Workspace, path: str):
        with open(path, "w") as f:
            toml.dump(active_workspace.export(), f)



from biscuit.common.ui import IconButton

from .view import View


class PanelView(View):
    """Abstract class for creating a panel view."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.__actions__ = []

    def add_action(self, icon, event=lambda *_: ...) -> None:
        self.__actions__.append((icon, event))

    def generate_actions(self, panelbar) -> None:
        self.__actions__ = [
            IconButton(
                panelbar, *action, **self.base.theme.layout.content.editors.bar.tab
            )
            for action in self.__actions__
        ]



from __future__ import annotations

import threading
import tkinter as tk
import typing
from tkinter.filedialog import askopenfilenames

from biscuit.common.chat import ChatModelInterface
from biscuit.common.icons import Icons
from biscuit.common.ui import Button, Entry, Frame, IconButton

from .attachments import Attachments
from .renderer import Renderer

if typing.TYPE_CHECKING:
    from .ai import AI


class Chat(Frame):
    """Chat view for the AI assistant.

    The Chat view is used to interact with the AI assistant.
    - The user can ask questions and get responses from the AI assistant.
    - The AI assistant is powered by the Google AI Studio API.
    - The user needs to enter the API key to start using the AI assistant.
    - The chat can be refreshed to start a new chat."""

    attachments: list[str] = []

    def __init__(self, master: AI, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: AI = master

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        theme = self.base.theme

        self.model = None

        container = Frame(self)
        container.grid(column=0, row=0, sticky=tk.NSEW)

        self.renderer = Renderer(container)
        self.renderer.pack(fill=tk.BOTH, expand=True)
        self.renderer.write("Hello! I'm Bikkis, How can I help you today?", True)

        self.entrybox = Frame(self, bg=theme.border)
        self.entrybox.grid(column=0, row=1, sticky=tk.EW)

        self.attachbtn = IconButton(self.entrybox, icon=Icons.ATTACH, event=self.attach)
        self.attachbtn.config(
            fg=theme.secondary_foreground, bg=theme.secondary_background, padx=5
        )
        self.attachbtn.pack(fill=tk.Y, pady=(1, 0), side=tk.LEFT)

        self.entry = Entry(self.entrybox, "Ask me anything...")
        self.entry.config(bg=theme.secondary_background)
        self.entry.bind("<Return>", self.send)
        self.entry.pack(fill=tk.BOTH, expand=True, side=tk.LEFT, pady=(1, 0))

        self.sendbtn = IconButton(self.entrybox, icon=Icons.SEND, event=self.send)
        self.sendbtn.config(fg=theme.biscuit, bg=theme.secondary_background, padx=5)
        self.sendbtn.pack(fill=tk.Y, pady=(1, 0))

        self.attached_indicator = Button(
            self,
            text="0 files attached",
        )
        self.attached_indicator.config(
            bg=theme.biscuit,
            fg=theme.secondary_background,
            font=self.base.settings.uifont,
        )
        self.attachments_window = Attachments(self)
        self.attached_indicator.set_command(self.attachments_window.show)

    def set_model(self, model: ChatModelInterface) -> None:
        self.model = model
        self.model.start_chat()

    def attach(self, *_) -> None:
        self.attachments += list(
            askopenfilenames(initialdir=self.base.active_directory)
        )
        if self.attachments:
            self.show_attachments()
        else:
            self.clear_attachments()

    def refresh_attachments(self) -> None:
        if self.attachments:
            self.show_attachments()
        else:
            self.clear_attachments()

    def attach_file(self, *files: typing.List[str]) -> None:
        if files:
            self.attachments += files
            self.show_attachments()

    def show_attachments(self) -> None:
        self.attached_indicator.grid(column=0, row=2, sticky=tk.EW)
        self.attached_indicator.config(text=f"{len(self.attachments)} files attached")

    def clear_attachments(self) -> None:
        self.attachments = []
        self.attached_indicator.grid_forget()

    def generate_response(self, question: str) -> str:
        try:
            for file in self.attachments:
                with open(file, "rb") as f:
                    question += f"\n\nContents of {file}:\n{f.read().decode()}"

            response = self.model.send_message(question)
            self.renderer.write(response, True)

            self.clear_attachments()
        except Exception as e:
            # self.base.notifications.error("Bikkis: Invalid API Key")
            # self.master.add_placeholder()
            self.renderer.write("Something went wrong. Please try again later", True)
            self.base.logger.error(e)

    def send(self, *_):
        text = self.entry.get()
        self.renderer.write(
            f"<p><font color={self.base.theme.biscuit}> You: "
            + text
            + "</font><br></p>"
            # show attachments
            # + "".join(
            #     [
            #         f"<div><font color={self.base.theme.biscuit}> Attached: {file}</font><br></div>"
            #         for file in self.attachments
            #     ]
            # ),
        )
        self.entry.delete(0, tk.END)

        threading.Thread(
            target=self.generate_response, args=(text,), daemon=True
        ).start()

    def new_chat(self) -> None:
        self.model.start_chat()
        self.renderer.content = ""
        self.renderer.write("Hello! I'm Bikkis, How can I help you today?", True)



from __future__ import annotations

import os
import typing

if typing.TYPE_CHECKING:
    from biscuit import App

    from .git import Git


class GitIgnore:
    def __init__(self, master):
        self.master: Git = master
        self.base: App = master.base
        self.path = ""
        self.repo = self.master.repo

    def load(self) -> None:
        if not self.base.git_found:
            return

        self.repo = self.master.repo
        self.path = os.path.join(self.base.active_directory, ".gitignore")

    def check(self, path: list[str]) -> list:
        """returns list of ignored files"""

        if not self.base.git_found:
            return []

        return self.repo.ignored(path)

    def add(self, path: str) -> None:
        """Add the given path to the .gitignore file."""

        if not self.base.git_found:
            return

        with open(self.path, "a") as f:
            f.write("\n" + path.replace("\\" "/"))

    def exclude(self, path: str) -> None:
        """Exclude the given path from the .gitignore file."""

        if not self.base.git_found:
            return

        with open(self.path, "a") as f:
            f.write("\n!" + path.replace("\\", "/"))



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Button, Frame, IconLabelButton, WrappingLabel


class DirectoryTreePlaceholder(Frame):
    """Placeholder view for the directory tree.

    The DirectoryTreePlaceholder is displayed when the user has not yet opened a folder.
    - The user can open a folder or clone a repository locally.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)
        self.columnconfigure(0, weight=1)

        WrappingLabel(
            self,
            text="No folder opened",
            font=self.base.settings.uifont,
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content
        ).grid(row=0, sticky=tk.EW)

        open_btn = Button(
            self,
            text="Open Folder",
            command=self.open_folder,
            pady=2,
        )
        open_btn.grid(row=1, pady=5, sticky=tk.EW)

        clone_btn = Button(
            self,
            text="Clone Repo",
            command=self.clone_repo,
            pady=2,
        )
        clone_btn.grid(row=3, sticky=tk.EW)

    def open_folder(self, *_) -> None:
        self.base.commands.open_directory()

    def clone_repo(self, *_) -> None:
        self.base.palette.show("clone:")



import os
import sys
from pathlib import Path

from .api import ExtensionsAPI
from .binder import Binder
from .commands import Commands
from .common import GameManager, SysInfo
from .debugger import DebuggerManager
from .execution import ExecutionManager
from .extensions import ExtensionManager
from .git import Git
from .history import HistoryManager
from .language import LanguageServerManager
from .session import SessionManager
from .settings import Settings
from .workspaces import WorkspaceManager


class ConfigManager:
    """
    CONFIG MANAGER
    --------------

    Configuration manager part of Biscuit Core.
    Manages the application state, configurations, settings, and extensions.
    """

    # application state / environement
    initialized: bool
    testing: bool

    # active directory dependant
    git_found: bool
    active_directory: str
    active_branch_name: str

    # constants
    resdir: str
    appdir: str
    extensiondir: str
    git_found = False
    wrap_words = False
    tab_spaces = 4
    relative_line_numbers = False
    block_cursor = False
    active_directory = None
    active_branch_name = None
    onupdate_callbacks = []
    onfocus_callbacks = []

    # runtime flags
    testing = False
    frozen = False

    def setup_configs(self) -> None:
        self.frozen = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
        self.testing = os.environ.get("ENVIRONMENT") == "test"

        self.system = SysInfo(self)
        self.settings = Settings(self)
        self.history = HistoryManager(self)
        self.sessions = SessionManager(self)
        self.workspaces = WorkspaceManager(self)

        self.resources = self.settings.resources
        self.bindings = self.settings.bindings
        self.config = self.settings.config
        self.theme = self.config.theme

        self.commands = Commands(self)
        self.binder = Binder(self)
        self.git = Git(self)
        self.game_manager = GameManager(self)
        self.language_server_manager = LanguageServerManager(self)
        self.execution_manager = ExecutionManager(self)
        self.debugger_manager = DebuggerManager(self)

    def setup_path(self, appdir: str) -> None:
        """Sets up the application paths and directories."""

        # sets up the application directory
        self.appdir = os.path.dirname(appdir)
        sys.path.append(self.appdir)

        # resources, config, extensions are outside src/
        if self.frozen:
            self.appdir = sys._MEIPASS
            self.parentdir = self.appdir
        else:
            self.appdir = os.path.dirname(os.path.abspath(__file__))
            self.parentdir = Path(self.appdir).parent.parent.absolute()

        self.configdir = os.path.join(self.parentdir, "config")

        self.extensiondir = Path.home() / ".biscuit" / "extensions"
        self.fallback_extensiondir = os.path.join(self.parentdir, "extensions")

        self.datadir = os.path.join(self.parentdir, "data")

        self.resdir = os.path.join(self.parentdir, "resources")
        self.fallback_resdir = Path(self.appdir).parent.absolute() / "resources"
        self.second_fallback_resdir = Path(self.appdir).absolute() / "resources"

        try:
            Path(self.datadir).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(f"Data directory invalid: {e}")

        try:
            Path(self.extensiondir).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(f"Extensions directory invalid: {e}")
            try:
                # fallback
                Path(self.fallback_extensiondir).mkdir(parents=True, exist_ok=True)
            except Exception as e:
                print(f"Extensions failed: {e}")

    def setup_extensions(self):
        # sets up the extension API & loads extensions
        self.api = ExtensionsAPI(self)
        self.extensions_manager = ExtensionManager(self)
        self.extensions_manager.start_server()
        self.extensions_view.results.late_setup()
        self.extensions_view.initialize()

    def set_tab_spaces(self, spaces: int) -> None:
        self.tab_spaces = spaces
        if e := self.editorsmanager.active_editor:
            if e.content and e.content.editable:
                e.content.text.set_tab_size(spaces)
                self.statusbar.set_spaces(spaces)

    @property
    def active_workspace(self):
        return self.workspaces.workspace



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Scrollbar

from ..panelview import PanelView


class Problems(PanelView):
    """The Problems view.

    The Problems view displays the problems detected by the application.
    - Show errors, warnings, and other problems.
    - Clear all problems.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__actions__ = ((Icons.CLEAR_ALL, self.clear),)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.text = tk.Text(
            self,
            relief=tk.FLAT,
            padx=10,
            pady=10,
            font=("Consolas", 11),
            **self.base.theme.editors.text
        )
        self.text.grid(row=0, column=0, sticky=tk.NSEW)

        self.scrollbar = Scrollbar(self, style="EditorScrollbar")
        self.scrollbar.grid(sticky=tk.NSEW, row=0, column=1)

        self.text.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.text.yview)

        self.text.tag_config("no-problems", foreground="gray")
        self.clear()

    def write(self, text) -> None:
        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.insert(tk.END, text)
        self.text.see(tk.END)
        self.text.config(state=tk.DISABLED)

    def clear(self, *_) -> None:
        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.insert(tk.END, "No problems detected.", "no-problems")
        self.text.config(state=tk.DISABLED)



from .endpoint import Endpoint


class Assistant(Endpoint):
    """Assistant endpoint

    This class is used to interact with the assistant API.
    """

    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.ai = self.base.ai

        self.register_provider = self.ai.register_provider



import tkinter as tk

from PIL import Image, ImageTk

from .editor import BaseEditor


class ImageViewer(BaseEditor):
    """ImageViewer is a simple image viewer that displays images in a canvas.
    User can zoom in and out using the mouse wheel and reset the zoom using Ctrl+0."""

    def __init__(self, master, path, editable=False, *args, **kwargs) -> None:
        super().__init__(master, path, editable=editable, *args, **kwargs)
        self.exists = True

        self.original_image = Image.open(self.path)
        # self.original_image.thumbnail((700, 700), Image.LANCZOS)
        self.display_image = self.original_image.copy()
        self.photo = ImageTk.PhotoImage(self.display_image)
        self.canvas = tk.Canvas(
            self,
            width=self.photo.width(),
            height=self.photo.height(),
            bd=0,
            highlightthickness=0,
            **self.base.theme.editors
        )
        self.canvas.pack(fill=tk.BOTH, expand=True)

        self.bind_all("<MouseWheel>", self.zoom_image)
        self.bind_all("<Control-0>", self.reset_zoom)
        self.after(100, self.create_image)

    def create_image(self):
        # print(self.canvas.winfo_width() // 2, self.canvas.winfo_height() // 2)
        self.canvas.create_image(
            self.canvas.winfo_width() // 2,
            self.canvas.winfo_height() // 2,
            anchor=tk.CENTER,
            image=self.photo,
        )

    def zoom_image(self, event: tk.Event):
        x, y = self.canvas.canvasx(event.x), self.canvas.canvasy(event.y)
        bbox = self.canvas.bbox(tk.ALL)
        if event.delta > 0:
            zoom_factor = 1.2
        else:
            zoom_factor = 1 / 1.2

        new_width = int(self.display_image.width * zoom_factor)
        new_height = int(self.display_image.height * zoom_factor)
        if new_width < 100 or new_height < 100 or new_width > 1000 or new_height > 1000:
            return

        new_image = self.original_image.copy().resize(
            (new_width, new_height), Image.LANCZOS
        )
        new_x = x - (x - bbox[0]) * zoom_factor
        new_y = y - (y - bbox[1]) * zoom_factor

        self.display_image = new_image
        self.photo = ImageTk.PhotoImage(self.display_image)
        self.canvas.delete("all")
        self.canvas.create_image(new_x, new_y, anchor=tk.NW, image=self.photo)

    def reset_zoom(self, event):
        self.display_image = self.original_image.copy()
        self.photo = ImageTk.PhotoImage(self.display_image)
        self.canvas.delete("all")
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.photo)



from __future__ import annotations

import os
import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Closable, Toplevel

if typing.TYPE_CHECKING:
    from .chat import Chat


class Attachments(Toplevel):
    """Attachments view for the AI assistant.

    The Attachments window is used to manage files attached to the AI chat.
    - The user can attach files to the AI assistant to get context based responses."""

    def __init__(self, master: Chat, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: Chat = master
        self.config(bg=self.base.theme.biscuit, padx=1, pady=1)

        self.visible = False

        self.overrideredirect(True)
        self.attachments: list[str] = []
        self.withdraw()

        self.bind("<FocusOut>", self.hide)

    def show(self, *_) -> None:
        """Show the attachments view."""

        self.visible = True

        self.update_idletasks()
        self.update()

        self.minsize(self.master.attached_indicator.winfo_width(), 0)

        for widget in self.winfo_children():
            widget.destroy()

        for file in self.master.attachments:
            item = Closable(
                self,
                text=os.path.basename(file),
                icon=Icons.FILE,
                fg=self.base.theme.biscuit,
                hfg=self.base.theme.biscuit_light,
            )
            item.close_btn.set_callback(
                lambda *_, item=item, file=file: self.remove_attachment(file, item)
            )
            item.text_label.config(anchor=tk.W)
            item.pack(fill=tk.X, expand=True)

        self.configure_geometry()
        self.deiconify()
        self.focus_set()

    def configure_geometry(self):
        self.update_idletasks()
        self.master.update_idletasks()

        x = self.master.attached_indicator.winfo_rootx()
        y = (
            self.master.attached_indicator.winfo_rooty()
            - self.winfo_height()
            - self.master.entrybox.winfo_height()
        )

        self.geometry(f"+{int(x)}+{int(y)}")

    def remove_attachment(self, file: str, item: Closable) -> None:
        """Remove an attachment from the attachments list."""

        self.master.attachments.remove(file)
        if not self.master.attachments:
            self.hide()

        item.destroy()
        self.master.refresh_attachments()
        self.configure_geometry()

    def hide(self, *_) -> None:
        """Hide the attachments view."""

        self.withdraw()
        self.visible = False



import tkinter as tk

from hintedtext import HintedEntry

from biscuit.common.ui import Frame


class Searchbar(Frame):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.config(bg=self.base.theme.biscuit)

        frame = Frame(self, **self.base.theme.editors)
        frame.pack(fill=tk.BOTH, padx=1, pady=1)

        self.text_variable = tk.StringVar()
        self.searchbar = HintedEntry(
            frame,
            font=self.base.settings.uifont,
            hint="Search settings",
            relief=tk.FLAT,
            textvariable=self.text_variable,
            **self.base.theme.editors.text
        )
        self.searchbar.pack(fill=tk.X, expand=True, pady=5, padx=5)

        self.configure_bindings()

    def configure_bindings(self) -> None:
        self.text_variable.trace("w", self.filter)

    def clear(self) -> None:
        self.text_variable.set("")

    def focus(self) -> None:
        self.searchbar.focus()

    def get_search_term(self) -> str:
        return self.searchbar.get().lower()

    def filter(self, *args) -> str:
        term = self.get_search_term()
        return
        new = [
            i for i in self.master.active_set if i[0].lower().startswith(term.lower())
        ]
        new += [
            i
            for i in self.master.active_set
            if any(
                [
                    f.lower() in i[0].lower()
                    or i[0].lower() in f.lower()
                    and i not in new
                    for f in term.lower().split()
                ]
            )
        ]

        self.master.show_result(new)



import os

__all__ = [
    "get_eol",
    "get_eol_label",
    "get_default_newline",
    "is_text_file",
    "eol_map",
    "eol_map_rev",
]

eol_map = {
    "LF": "\n",
    "CRLF": "\r\n",
    "CR": "\r",
}
eol_map_rev = dict((v, k) for k, v in eol_map.items())


def get_eol(newline):
    """Get the EOL character for the given newline setting."""

    return eol_map.get(newline, os.linesep)


def get_eol_label(newline):
    """Get the label for the given newline setting."""

    return eol_map_rev.get(newline, "AUTO")


def get_default_newline():
    """Get the default newline setting."""

    return os.linesep


# CREDITS: https://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python/
int2byte = lambda x: bytes((x,))
_text_characters = b"".join(int2byte(i) for i in range(32, 127)) + b"\n\r\t\f\b"


def is_text_file(path: str) -> bool:
    """Check if a file is a text file.

    Args:
        path (str): path to the file
    """

    fp = open(path, "rb")
    block = fp.read(512)
    fp.close()

    if b"\x00" in block:
        return False
    elif not block:
        return True

    nontext = block.translate(None, _text_characters)
    return float(len(nontext)) / len(block) <= 0.30



import tkinter as tk

from ..icons import Icons
from ..ui.icon import IconButton
from ..ui.native import Frame
from .menu import Menu


class _DropdownMenu(Menu):
    def get_coords(self, e) -> tuple:
        return (
            self.master.winfo_rootx(),
            self.master.winfo_rooty() + self.master.winfo_height(),
        )


class Dropdown(Frame):
    """For implementing a dropdown menu."""

    def __init__(
        self,
        master,
        selected="",
        items=[],
        icon: Icons = "",
        callback=lambda *_: None,
        iconside=tk.LEFT,
        padx=5,
        pady=1,
        fg="",
        bg="",
        hfg="",
        hbg="",
        iconfg="",
        iconbg="",
        iconhfg="",
        iconhbg="",
        empty_message="No items",
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, padx=padx, pady=pady, *args, **kwargs)
        self.callback = callback
        self.empty_message = empty_message

        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.utils.iconlabelbutton.values()
        )
        if fg:
            self.fg = fg
        if bg:
            self.bg = bg
        if hfg:
            self.hfg = hfg
        if hbg:
            self.hbg = hbg

        self.iconfg = iconfg or self.fg
        self.iconbg = iconbg or self.bg
        self.iconhfg = iconhfg or self.hfg
        self.iconhbg = iconhbg or self.hbg

        self.config(bg=self.bg)
        self.text = selected or empty_message
        self.icon = icon

        self.icon_label = None
        self.text_label = None

        self.selected = None
        self.menu = _DropdownMenu(self)
        self.set_items(items)

        if icon:
            self.icon_label = tk.Label(
                self,
                text=self.icon,
                anchor=tk.CENTER,
                bg=self.iconbg,
                fg=self.iconfg,
                font=("codicon", 12),
            )
            self.icon_label.pack(side=iconside, fill=tk.BOTH)

        self.text_label = tk.Label(
            self,
            text=self.text,
            anchor=tk.CENTER,
            pady=2,
            bg=self.bg,
            fg=self.fg,
            font=self.base.settings.uifont,
        )
        self.text_label.pack(side=iconside, fill=tk.BOTH, expand=True)

        self.dropdown_btn = IconButton(self, Icons.CHEVRON_DOWN, self.menu.show)
        self.dropdown_btn.config(bg=self.bg, fg=self.fg)
        self.dropdown_btn.pack(side=tk.RIGHT, fill=tk.BOTH)

        self.config_bindings()
        self.visible = False

    def add_command(self, text) -> None:
        """Add a command to the dropdown menu"""

        self.menu.add_command(text, lambda text=text: self.choose(text))

    def set_items(self, items: list[str]) -> None:
        self.menu.clear()
        if not items:
            self.add_command(self.empty_message)
        else:
            for i in items:
                self.add_command(i)

    def config_bindings(self) -> None:
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

        self.bind("<Button-1>", self.menu.show)
        if self.text:
            self.text_label.bind("<Button-1>", self.menu.show)
        if self.icon:
            self.icon_label.bind("<Button-1>", self.menu.show)

    def on_enter(self, *_) -> None:
        self.config(bg=self.hbg)
        if self.text:
            self.text_label.config(bg=self.hbg, fg=self.hfg)
        if self.icon:
            self.icon_label.config(bg=self.iconhbg, fg=self.iconhfg)
        self.dropdown_btn.config(bg=self.hbg, fg=self.hfg)

    def on_leave(self, *_) -> None:
        self.config(bg=self.bg)
        if self.text:
            self.text_label.config(bg=self.bg, fg=self.fg)
        if self.icon:
            self.icon_label.config(bg=self.iconbg, fg=self.iconfg)
        self.dropdown_btn.config(bg=self.bg, fg=self.fg)

    def change_text(self, text) -> None:
        """Change the text of the item"""

        self.text_label.config(text=text)

    def change_icon(self, icon) -> None:
        """Change the icon of the item"""

        self.icon_label.config(text=icon)

    def set_pack_data(self, **kwargs) -> None:
        self.pack_data = kwargs

    def get_pack_data(self):
        return self.pack_data

    def choose(self, text) -> None:
        """Choose an item from the dropdown menu"""

        self.selected = text
        self.text_label.config(text=text)
        self.callback(text)

    def show(self) -> None:
        """Show the item"""

        if not self.visible:
            self.visible = True
            self.pack(**self.get_pack_data())

    def hide(self) -> None:
        """Hide the item"""

        if self.visible:
            self.visible = False
            self.pack_forget()
